from samson.oracles.encryption_oracle import EncryptionOracle

class StatelessBlockEncryptionOracle(EncryptionOracle):
    """
    Oracle that provides an interface to a chosen-plaintext attack on a stateless block encryption.
    """

    def find_block_size(self) -> int:
        """
        Attempts to figure out the cipher's block size by probing the oracle.

        Returns:
            int: The cipher's block size.
        """
        baseline = len(self.encrypt(b''))

        for i in range(33):
            curr_len = len(self.encrypt(('A' * i).encode()))
            if  curr_len != baseline:
                return curr_len - baseline


from samson.block_ciphers.rijndael import Rijndael
from samson.utilities.bytes import Bytes

class Poly1305(object):
    """
    Message authentication code using an underlying block cipher. The (r, nonce) combination MUST
    be unique to guarantee its security properties. A single reuse can allow for a forgery.
    """

    P1305 = (1 << 130) - 5
    
    def __init__(self, key: bytes, nonce: bytes, r: bytes, cipher=Rijndael):
        """
        Parameters:
            key    (bytes): Bytes-like object to key the underlying cipher.
            nonce  (bytes): Bytes-like nonce.
            r      (bytes): Bytes-like polynomial.
            cipher (class): Instantiable class representing a block cipher.
        """
        self.key = key
        self.nonce = nonce
        self.r = Bytes.wrap(r, byteorder='little').to_int()
        self.cipher = cipher


    def __repr__(self):
        return f"<Poly1305: key={self.key}, nonce={self.nonce}, r={self.r}, cipher={self.cipher}>"


    def __str__(self):
        return self.__repr__()

    
    def generate(self, message: bytes) -> Bytes:
        """
        Generates a keyed MAC for `message`.

        Parameters:
            message (bytes): Message to generate a MAC for.
        
        Returns:
            Bytes: The MAC.
        """
        pt_chunks = [(chunk + b'\x01').zfill(17) for chunk in Bytes.wrap(message, byteorder='little').chunk(16, allow_partials=True)]

        total = 0
        for chunk in pt_chunks:
            total += chunk.to_int()
            total *= self.r
        
        total %= Poly1305.P1305
        enc = self.cipher(self.key).encrypt(self.nonce)
        enc.byteorder = 'little'

        return Bytes((enc.to_int() + total) % (1 << 128), byteorder='little')


from math import ceil

def xor_buffs(buf1: bytes, buf2: bytes) -> bytearray:
    """
    XORs two byte buffers.

    Parameters:
        buf1 (bytes): First byte buffer.
        buf2 (bytes): Second byte buffer.
    
    Returns:
        bytearray: Resulting bytes.
    """
    if len(buf1) != len(buf2):
        raise Exception('Buffers must be equal length.')

    return bytearray([x ^ y for x, y in zip(buf1, buf2)])



def stretch_key(key: bytes, length: int, offset: int=0) -> bytes:
    """
    Repeats a bytes object until it reaches `size` length shifted by `offset`.

    Examples:

    >>> stretch_key(b'abc', 5)
    b'abcab'

    >>> stretch_key(b'abc', 5, offset=1)
    b'cabca'


    Parameters:
        size   (int): Size to be stretched to.
        offset (int): Offset to start from.
    
    Returns:
        bytes: Bytes stretched to `size`.
    """
    offset_mod = offset % len(key)
    key_stream = (key * ceil(length / len(key)))
    complete_key = (key[-offset_mod:] + key_stream)[:length]
    return complete_key



def transpose(ciphertext: bytes, size: int) -> bytes:
    """
    Builds a matrix of `size` row-length, transposes the matrix, and collapses it back into a bytes object.

    Parameters:
        size (int): Length of the rows/chunks.
    
    Returns:
        bytes: Transposed bytes.
    """
    return [ciphertext[i::size] for i in range(size)]



def get_blocks(ciphertext: bytes, block_size: int=16, allow_partials: bool=False) -> list:
    """
    Chunks the bytes into `size` length chunks.

    Parameters:
        size            (int): Size of the chunks.
        allow_partials (bool): Whether or not to allow the last chunk to be a partial.
    
    Returns:
        list: List of bytes.
    """
    full_blocks = [ciphertext[i * block_size: (i + 1) * block_size] for i in range(len(ciphertext) // block_size)]

    left_over = len(ciphertext) % block_size

    if allow_partials and left_over > 0:
        all_blocks = full_blocks + [ciphertext[-left_over:]]
    else:
        all_blocks = full_blocks
    return all_blocks



def left_rotate(x: int, amount: int, bits: int=32) -> int:
    """
    Performs a left-rotate.

    Parameters:
        x      (int): Integer to rotate.
        amount (int): Amount to rotate by.
        bits   (int): Bitspace to rotate over.
    
    Returns:
        int: Rotated integer.
    """
    mask = 2 ** bits - 1
    x &= mask
    return ((x<<amount) | (x>>(bits-amount))) & mask



def right_rotate(x: int, amount: int, bits: int=32) -> int:
    """
    Performs a right-rotate.

    Parameters:
        x      (int): Integer to rotate.
        amount (int): Amount to rotate by.
        bits   (int): Bitspace to rotate over.
    
    Returns:
        int: Rotated integer.
    """
    mask = 2 ** bits - 1
    return ((x>>amount) | (x<<(bits-amount))) & mask


from samson.utilities.math import mod_inv
from samson.utilities.bytes import Bytes
from samson.publickey.dsa import DSA
from fastecdsa.point import Point

# https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm
class ECDSA(DSA):
    """
    Elliptical Curve Digital Signature Algorithm
    """

    def __init__(self, G: Point, hash_obj: object, d: int=None):
        """
        Parameters:
            G         (Point): Generator point for a curve.
            hash_obj (object): Instantiated object with compatible hash interface.
            d           (int): (Optional) Private key.
        """
        self.G = G
        self.q = self.G.curve.q
        self.d = d or max(1, Bytes.random(self.q.bit_length() + 7 // 8).int() % self.q)
        self.Q = self.d * self.G
        self.hash_obj = hash_obj

    
    def __repr__(self):
        return f"<ECDSA: d={self.d}, G={self.G}, Q={self.Q}, hash_obj={self.hash_obj}>"


    def __str__(self):
        return self.__repr__()
    

    def sign(self, message: bytes, k: int=None) -> (int, int):
        """
        Signs a `message`.

        Parameters:
            message (bytes): Message to sign.
            k         (int): (Optional) Ephemeral key.
        
        Returns:
            (int, int): Signature formatted as (r, s).
        """
        r = 0
        s = 0

        while s == 0 or r == 0:
            k = k or max(1, Bytes.random(self.q .bit_length() + 7 // 8).int() % self.q)
            inv_k = mod_inv(k, self.q)

            z = self.hash_obj.hash(message).int()
            z >>= max(self.hash_obj.digest_size * 8 - self.q.bit_length(), 0)

            r = (k * self.G).x % self.q
            s = (inv_k * (z + self.d * r)) % self.q

        return (r, s)
    
    
    def verify(self, message: bytes, sig: (int, int)) -> bool:
        """
        Verifies a `message` against a `sig`.

        Parameters:
            message  (bytes): Message.
            sig ((int, int)): Signature of `message`.
        
        Returns:
            bool: Whether the signature is valid or not.
        """
        (r, s) = sig
        w = mod_inv(s, self.q)

        z = self.hash_obj.hash(message).int()
        z >>= max(self.hash_obj.digest_size * 8 - self.q.bit_length(), 0)

        u_1 = (z * w) % self.q
        u_2 = (r * w) % self.q
        v = u_1 * self.G + u_2 * self.Q
        return v.x == r


import math
from samson.macs.hmac import HMAC
from samson.utilities.bytes import Bytes

# https://en.wikipedia.org/wiki/HKDF
# https://tools.ietf.org/html/rfc5869
class HKDF(object):
    """
    Key derivation function based on HMAC. Formally described in RFC5869 (https://tools.ietf.org/html/rfc5869).
    """

    def __init__(self, hash_obj: object, desired_len: int):
        """
        Parameters:
            hash_obj (object): Instantiated object with compatible hash interface.
            desired_len (int): Desired output length (in bytes).
        """
        self.hash_obj = hash_obj
        self.desired_len = desired_len


    def __repr__(self):
        return f"<HKDF: hash_obj={self.hash_obj}, desired_len={self.desired_len}>"

    def __str__(self):
        return self.__repr__()

    
    def derive(self, key: bytes, salt: bytes, info: bytes=b'') -> Bytes:
        """
        Derives a key.

        Parameters:
            key  (bytes): Bytes-like object to key the internal HMAC.
            salt (bytes): Salt to tweak the output.
            info (bytes): Additional data to use as tweak.
        
        Returns:
            Bytes: Derived key.
        """
        prk = HMAC(key=salt, hash_obj=self.hash_obj).generate(key)
        hmac = HMAC(key=prk, hash_obj=self.hash_obj)

        new_key = b''
        t = b''
        for i in range(math.ceil(self.desired_len / (self.hash_obj.digest_size // 8))):
            t = hmac.generate(t + info + bytes([i + 1]))
            new_key += t

        return new_key[:self.desired_len]


from sympy import GF
from samson.utilities.math import mod_inv
from fastecdsa.curve import P192, P224, P256, P384, P521
import math

# Convenience aliases
secp192r1 = P192
secp224r1 = P224
secp256r1 = P256
secp384r1 = P384
secp521r1 = P521


# https://tools.ietf.org/html/rfc7748
def cswap(swap: int, x_2: int, x_3: int) -> (int, int):
    """
    Conditional constant-time swap.

    Parameters:
        swap (int): 0 or 1. 1 means swap.
        x_2  (int): First int.
        x_3  (int): Second int.

    Returns:
        (int, int): Formatted as (x_2, x_3)
    """
    dummy = swap * (x_2 - x_3)
    x_2 = x_2 - dummy
    x_3 = x_3 + dummy
    return (x_2, x_3)



# https://tools.ietf.org/html/rfc7748#section-4.1
class MontgomeryCurve(object):
    """
    Montgomery Curve

    Basically just decouples parameters from `MontgomeryPoint`.
    """

    # # https://tools.ietf.org/html/rfc7748#section-3
    def __init__(self, p: int, A: int, a24: int, U: int, V: int):
        """
        Parameters:
            p   (int): Prime number defining the underlying field.
            A   (int): An element in the finite field GF(p), not equal to -2 or 2.
            a24 (int): Constant for curve multiplication.
            U   (int): The u-coordinate of the elliptic curve point P on a Montgomery curve.
            V   (int): The v-coordinate of the elliptic curve point P on a Montgomery curve.
        """
        self.p = p
        self.A = A
        self.a24 = a24
        self.U = U
        self.V = V


    def __repr__(self):
        return f"<MontgomeryCurve: p={self.p}, A={self.A}, U={self.U}, V={self.V}>"

    def __str__(self):
        return self.__repr__()


    def __eq__(self, other) -> bool:
        return self.p == other.p and self.A == other.A and self.U == other.U and self.V == other.V

    


class MontgomeryPoint(object):
    """
    Point on a Montgomery Curve

    Provides scalar multiplication.
    """

    def __init__(self, x: int, curve: MontgomeryCurve):
        """
        Parameters:
            x                 (int): x-coordinate.
            curve (MontgomeryCurve): The underlying curve.
        """
        self.curve = curve
        self.x = x


    def __repr__(self):
        return f"<MontgomeryPoint: x={self.x}, curve={self.curve}>"

    def __str__(self):
        return self.__repr__()

    

    def __eq__(self, other) -> bool:
        return self.x == other.x and self.curve == other.curve


    # https://tools.ietf.org/html/rfc7748#section-5
    def __mul__(self, other: int) -> int:
        if type(other) != int:
            raise NotImplementedError("MontgomeryPoint multiplication is currently only implemented for scalars.")

        A = self.curve.A
        gf_p = GF(self.curve.p)

        x_1 = gf_p(other)
        x_2 = gf_p(1)
        z_2 = gf_p(0)
        x_3 = gf_p(other)
        z_3 = gf_p(1)
        swap = 0

        for t in range(self.x.bit_length()-1, -1, -1):
            k_t = (self.x >> t) & 1
            swap ^= k_t

            # Conditional swap
            x_2, x_3 = cswap(swap, x_2, x_3)
            z_2, z_3 = cswap(swap, z_2, z_3)
            swap = k_t

            A = x_2 + z_2
            AA = A**2
            B = x_2 - z_2
            BB = B**2
            E = AA - BB
            C = x_3 + z_3
            D = x_3 - z_3
            DA = D * A
            CB = C * B
            x_3 = (DA + CB)**2
            z_3 = x_1 * (DA - CB)**2
            x_2 = AA * BB
            z_2 = E * (AA + self.curve.a24 * E)

        # Conditional swap
        x_2, x_3 = cswap(swap, x_2, x_3)
        z_2, z_3 = cswap(swap, z_2, z_3)
        return int(x_2 * (z_2**(self.curve.p - 2))) % self.curve.p

    
    def __rmul__(self, other):
        return self.__mul__(other)




class Curve25519(MontgomeryCurve):
    def __init__(self):
        super().__init__(p=2**255 - 19, A=486662, a24=121665, U=9, V=1478161944758954479102059356840998688726460613461647528896488183775558623740)
    

    def clamp_to_curve(self, x: int) -> MontgomeryPoint:
        """
        Coerces `x` to a valid x-coordinate on Curve25519.

        Parameters:
            x (int): `x` value to coerce.

        Returns:
            MontgomeryPoint: Valid MontgomeryPoint.
        """
        x &= ~7
        x &= ~(128 << 8 * 31)
        x |= 64 << 8 * 31
        return MontgomeryPoint(x, self)



class Curve448(MontgomeryCurve):
    def __init__(self):
        super().__init__(p=2**448 - 2**224 - 1, A=156326, a24=39081, U=5, V=355293926785568175264127502063783334808976399387714271831880898435169088786967410002932673765864550910142774147268105838985595290606362)
    

    def clamp_to_curve(self, x: int) -> MontgomeryPoint:
        """
        Coerces `x` to a valid x-coordinate on Curve448.

        Parameters:
            x (int): `x` value to coerce.

        Returns:
            MontgomeryPoint: Valid MontgomeryPoint.
        """
        x &= ~3
        x |= 128 << 8 * 55
        return MontgomeryPoint(x, self)



# https://ed25519.cr.yp.to/python/ed25519.py
# https://tools.ietf.org/html/rfc8032
class TwistedEdwardsCurve(object):
    """
    Twisted Edwards Curve

    Provides general curve operations and parameter decoupling.
    """

    # https://tools.ietf.org/html/rfc8032#section-5
    # https://tools.ietf.org/html/rfc8032#section-3
    def __init__(self, a: int, c: int, n: int, b: int, magic: bytes, q: int, l: int, d: int, B: (int, int)):
        """
        Parameters:
            a        (int): Twist parameter. a=1 is untwisted, the special case.
            c        (int): Base 2 logarithm of cofactor
            n        (int): Defines the number of bits in EdDSA scalars.
            b        (int): Number of bits the curve can encode.
            magic  (bytes): The magic byte-string (if any) of the curve.
            q        (int): Modulus.
            l        (int): Order of the curve.
            d        (int): A non-zero element in the finite field GF(p), not equal to 1, in the case of an Edwards curve, or not equal to -1, in the case of a twisted Edwards curve
            B ((int, int)): Base point.
        """
        self.a = a
        self.c = c
        self.n = n
        self.b = b
        self.magic = magic
        self.q = q
        self.l = l
        self.d = d
        self.I = pow(2, (q-1) // 4, q)
        self.B = TwistedEdwardsPoint(*B, self)


    def __repr__(self):
        return f"<TwistedEdwardsCurve: b={self.b}, q={self.q}, l={self.l}>"

    def __str__(self):
        return self.__repr__()

    

    def __eq__(self, other) -> bool:
        return self.b == other.b and self.q == other.q and self.l == other.l and self.d == other.d

    

    def is_on_curve(self, P: (int, int)) -> bool:
        """
        Determines if the point `P` is on the curve.

        Parameters:
            P (int, int): The point formatted as (x, y).

        Returns:
            bool: Whether or not the point is on the curve.
        """
        x, y = P
        return (self.a * x*x + y*y - 1 - self.d * x*x*y*y) % self.q == 0
    


    def recover_point_from_y(self, y: int):
        """
        Recovers the full TwistedEdwardsPoint from the y-coordinate.

        Parameters:
            y (int): y-coordinate of a valid TwistedEdwardsPoint.
        
        Returns:
            TwistedEdwardsPoint: Full TwistedEdwardsPoint.

        """
        xx = (y*y-1) * mod_inv(self.d*y*y-self.a, self.q)
        if self.q % 8 == 5:
            x = pow(xx, (self.q+3)//8, self.q)

            if (x*x - xx) % self.q != 0:
                x = (x*self.I) % self.q

            if x % 2 != 0:
                x = self.q-x

        elif self.q % 4 == 3:
            x = pow(xx, (self.q+1)//4, self.q)
        else:
            raise Exception("Unsupported prime `q`.")

        return TwistedEdwardsPoint(x % self.q, y % self.q, self)




class TwistedEdwardsPoint(object):
    """
    Point on a Twisted Edwards Curve

    Provides scalar multiplication and point addition.
    """

    def __init__(self, x: int, y: int, curve: TwistedEdwardsCurve, validate: bool=True):
        """
        Parameters:
            x                     (int): x-coordinate.
            y                     (int): y-coordinate.
            curve (TwistedEdwardsCurve): Underlying curve.
            validate             (bool): Whether or not to validate the point against the curve.
        """
        self.curve = curve
        self.x = x % curve.q
        self.y = y % curve.q

        if validate and not curve.is_on_curve((x, y)):
            raise ValueError(f"({x}, {y}) is not on {curve}")



    def __repr__(self):
        return f"<TwistedEdwardsPoint: x={self.x}, y={self.y}, curve={self.curve}>"

    def __str__(self):
        return self.__repr__()
    


    def __eq__(self, other) -> bool:
        return self.x == other.x and self.y == other.y and self.curve == other.curve


    def __add__(self, other):
        if type(other) != TwistedEdwardsPoint:
            raise TypeError("TwistedEdwardsPoint addition only defined between points.")
        
        assert self.curve == other.curve

        x1, y1 = self.x, self.y
        x2, y2 = other.x, other.y
        x3 = (x1*y2+x2*y1) * mod_inv(1+self.curve.d * x1*x2*y1*y2, self.curve.q)
        y3 = (y1*y2 - self.curve.a*x1*x2) * mod_inv(1-self.curve.d * x1*x2*y1*y2, self.curve.q)

        return TwistedEdwardsPoint(x3, y3, self.curve)
    

    def __mul__(self, scalar: int):
        if type(scalar) != int:
            raise NotImplementedError("TwistedEdwardsPoint multiplication is currently only implemented for scalars.")

        Q = TwistedEdwardsPoint(0, 1, self.curve)
        iters = math.ceil(math.log(scalar, 2))

        for i in range(iters-1, -1, -1):
            Q = Q + Q

            if (scalar >> i) & 1:
                Q = Q + self
        
        return Q


    def __rmul__(self, scalar: int):
        return self.__mul__(scalar)




EdwardsCurve25519 = TwistedEdwardsCurve(a=-1, c=3, n=254, b=256, magic=b'', q=2**255 - 19, l=2**252 + 27742317777372353535851937790883648493, d=-121665 * pow(121666, 2**255 - 19 -2, 2**255 - 19), B=(15112221349535400772501151409588531511454012693041857206046113283949847762202, 46316835694926478169428394003475163141307993866256225615783033603165251855960))
EdwardsCurve448   = TwistedEdwardsCurve(a=1, c=2, n=447, b=456, magic=b'SigEd448\x00\x00', q=2**448 - 2**224 - 1, l=2**446 - 0x8335dc163bb124b65129c96fde933d8d723a70aadc873d6d54a7bb0d, d=-39081, B=(224580040295924300187604334099896036246789641632564134246125461686950415467406032909029192869357953282578032075146446173674602635247710, 298819210078481492676017930443930673437544040154080242095928241372331506189835876003536878655418784733982303233503462500531545062832660))


from samson.utilities.math import modexp
from samson.utilities.bytes import Bytes
from samson.hashes.sha2 import SHA256


# http://srp.stanford.edu/design.html
# https://bnetdocs.org/document/24/nls-srp-protocol
# https://tools.ietf.org/html/rfc5054#
class SRPClient(object):
    """
    Secure Remote Password protocol client
    """

    MODP_1024 = 0xEEAF0AB9ADB38DD69C33F80AFA8FC5E86072618775FF3C0B9EA2314C9C256576D674DF7496EA81D3383B4813D692C6E0E0D5D8E250B98BE48E495C1D6089DAD15DC7D7B46154D6B6CE8EF4AD69B15D4982559B297BCF1885C529F566660E57EC68EDBC3C05726CC02FD4CBF4976EAA9AFD5138FE8376435B9FC61D2FC0EB06E3
    MODP_1536 = 0x9DEF3CAFB939277AB1F12A8617A47BBBDBA51DF499AC4C80BEEEA9614B19CC4D5F4F5F556E27CBDE51C6A94BE4607A291558903BA0D0F84380B655BB9A22E8DCDF028A7CEC67F0D08134B1C8B97989149B609E0BE3BAB63D47548381DBC5B1FC764E3F4B53DD9DA1158BFD3E2B9C8CF56EDF019539349627DB2FD53D24B7C48665772E437D6C7F8CE442734AF7CCB7AE837C264AE3A9BEB87F8A2FE9B8B5292E5A021FFF5E91479E8CE7A28C2442C6F315180F93499A234DCF76E3FED135F9BB
    MODP_2048 = 0xAC6BDB41324A9A9BF166DE5E1389582FAF72B6651987EE07FC3192943DB56050A37329CBB4A099ED8193E0757767A13DD52312AB4B03310DCD7F48A9DA04FD50E8083969EDB767B0CF6095179A163AB3661A05FBD5FAAAE82918A9962F0B93B855F97993EC975EEAA80D740ADBF4FF747359D041D5C33EA71D281E446B14773BCA97B43A23FB801676BD207A436C6481F1D2B9078717461A5B9D32E688F87748544523B524B0D57D5EA77A2775D2ECFA032CFBDBF52FB3786160279004E57AE6AF874E7303CE53299CCC041C7BC308D82A5698F3A8D0C38271AE35F8E9DBFBB694B5C803D89F7AE435DE236D525F54759B65E372FCD68EF20FA7111F9E4AFF73


    def __init__(self, identity: bytes, password: bytes, g: int=2, N: int=MODP_1024, hash_obj: object=SHA256(), a: int=None):
        """
        Parameters:
            identity  (bytes): Username.
            password  (bytes): Password.
            g           (int): Generator.
            N           (int): Prime modulus.
            hash_obj (object): Instantiated object with compatible hash interface.
            a           (int): Random private value.
        """
        self.a = a or Bytes.random(4).int() % N
        self.g = g
        self.A = modexp(g, self.a, N)
        self.identity = identity
        self.password = password
        self.N = N
        self.hash_obj = hash_obj
        self.k = hash_obj.hash(Bytes(N) + self.PAD(g)).int()



    def PAD(self, in_bytes: bytes) -> Bytes:
        """
        If a conversion is explicitly specified with the operator PAD(), the integer will first be implicitly converted, then the resultant byte-string will be left-padded with zeros (if necessary) until its length equals the implicitly-converted length of N.
        """
        return Bytes.wrap(in_bytes).zfill((self.N.bit_length() + 7) // 8)

    

    def make_request(self) -> (bytes, int):
        """
        Creates the initial client request.

        Returns:
            (bytes, int): Formatted as (identity, client's challenge `A`).
        """
        return self.identity, self.A



    def perform_challenge(self, salt: bytes, B: int) -> Bytes:
        """
        Performs server challenge.

        Parameters:
            salt  (bytes): Salt from server.
            B       (int): Server's challenge.
        
        Returns:
            Bytes: Challenge bytes to send to server.
        """
        uH = self.hash_obj.hash(self.PAD(self.A) + self.PAD(B)).int()
        xH = self.hash_obj.hash(salt + self.hash_obj.hash(self.identity + b':' + self.password)).int()

        p1 = (B - self.k * modexp(self.g, xH, self.N))
        p2 = (self.a + uH * xH)
        cS = modexp(p1, p2, self.N)

        cK = self.hash_obj.hash(Bytes(cS))
        return self.hash_obj.hash(cK + salt)



    def craft_malicious_request(self) -> (bytes, int):
        """
        Crafts a malicious request by setting the initial challenge parameter to zero.

        Returns:
            (bytes, int): Formatted as (identity, 0).
        """
        return self.identity, 0



    def craft_auth_bypass(self, salt: bytes) -> Bytes:
        """
        Crafts the authentication bypass challenge. Only works if server accepted the initial malicious request.

        Parameters:
            salt (bytes): Salt from server.

        Returns:
            Bytes: Challenge bytes to send to server.
        """
        cK = self.hash_obj.hash(b'\x00')
        return self.hash_obj.hash(cK + salt)


from samson.auxiliary.tokenizer_handler import TokenizerHandler

class TokenListHandler(TokenizerHandler):
    """
    TokenizerHandler that builds a list.
    """

    def __init__(self):
        self.list = []


    def reset(self):
        pass

    
    def handle_token(self, token: str):
        """
        Handles tokens found by a Tokenizer.

        Parameters:
            token (str): Token found.
        """
        self.list.append(token)


    def finalize(self) -> list:
        """
        Returns the list of tokens.
        """
        return self.list


from .a51 import A51
from .chacha import ChaCha
from .e0 import E0
from .rc4 import RC4
from .salsa import Salsa
from .snow3g import SNOW3G
from .zuc import ZUC


__all__ = ["A51", "ChaCha", "E0", "RC4", "Salsa", "SNOW3G", "ZUC"]


from samson.utilities.bitstring import Bitstring

class BitslicedFLFSR(object):
    """
    An implementation of an FLFSR using an internal bitstring rather than an integer.
    """
    
    def __init__(self, length: int, clock_bit: int, taps: list, seed: int=0):
        """
        Parameters:
            length    (int): Length of the internal, bitstring state.
            clock_bit (int): Bit considered to be the 'clock' bit.
            taps     (list): Positional taps used to determine the next bit.
            seed      (int): Inital state represented as an integer.
        """
        self.state = bin(seed)[2:].zfill(length)
        self.length = length
        self.clock_bit = clock_bit
        self.taps = taps


    def __repr__(self):
        return f"<BitslicedFLFSR: state={self.state}, length={self.length}, clockbit={self.clock_bit}, taps={self.taps}>"


    def __str__(self):
        return self.__repr__()


    def mix_state(self, in_val: bytes, size: int):
        """
        Mixes state into the FLSFR by clocking in each bit.

        Parameters:
            in_val (bytes): Bytes-like value to clock in.
            size     (int): Size for the bitstring to be padded to.
        """
        bit_val = Bitstring.wrap(in_val).zfill(size)

        for bit in str(bit_val)[::-1]:
            self.clock(int(bit))


    def clock(self, bit: int=0):
        """
        Clocks the FLSFR with an optional input value.

        Parameters:
            bit (int): (Optional) Value to XOR'd in with the next bit.
        """
        for tap in self.taps:
            position = self.length - tap - 1
            bit ^= int(self.state[position])
        
        self.state += '0'
        self.state = self.state[1:]
        self.state = (self.state[:self.length - 1] + str(bit))


    def value(self) -> int:
        """
        Retrives the current value in state[0].

        Returns:
            int: Current value.
        """
        return int(self.state[0])

    
    def clock_value(self) -> int:
        """
        Retrives the value in state[clock_bit].

        Returns:
            int: Clock value.
        """
        return int(self.state[self.clock_bit])


    
    def generate(self) -> int:
        """
        Calls self.clock(). Here for interface uniformity.

        Returns:
            int: Current value.
        """
        self.clock()
        return self.value()


from .blowfish import Blowfish
from .camellia import Camellia
from .des import DES
from .kasumi import KASUMI
from .rc5 import RC5
from .rijndael import Rijndael
from .serpent import Serpent
from .tdes import TDES
from .twofish import Twofish


__all__ = ["Blowfish", "Camellia", "DES", "KASUMI", "RC5", "Rijndael", "Serpent", "TDES", "Twofish"]


from .cbc_iv_key_equivalence_attack import CBCIVKeyEquivalenceAttack
from .cbc_padding_oracle_attack import CBCPaddingOracleAttack
from .compression_ratio_side_channel_attack import CompressionRatioSideChannelAttack
from .crime_attack import CRIMEAttack
from .ecb_prepend_attack import ECBPrependAttack
from .iterated_hash_multicollision import IteratedHashMulticollisionAttack
from .mangers_attack import MangersAttack
from .nostradamus_attack import NostradamusAttack
from .pkcs1v15_padding_oracle_attack import PKCS1v15PaddingOracleAttack
from .rc4_prepend_attack import RC4PrependAttack
from .xor_bitflipping_attack import XORBitflippingAttack
from .xor_dictionary_attack import XORDictionaryAttack
from .xor_transposition_attack import XORTranspositionAttack


__all__ = ["CBCIVKeyEquivalenceAttack", "CBCPaddingOracleAttack", "CompressionRatioSideChannelAttack", "CRIMEAttack", "ECBPrependAttack", "IteratedHashMulticollisionAttack", "MangersAttack", "NostradamusAttack", "PKCS1v15PaddingOracleAttack", "RC4PrependAttack", "XORBitflippingAttack", "XORDictionaryAttack", "XORTranspositionAttack"]


from math import ceil
import struct

class IteratedHashMulticollisionAttack(object):
    def __init__(self, hash_func):
        self.hash_func = hash_func
        self._collisions = {}


    def _find_collision(self, iv):
        hash_dict = {}

        for i in range(2**64):
            attempt = struct.pack('Q', i)
            digest = self.hash_func(iv, attempt)
            if digest in hash_dict:
                if (iv, digest) not in self._collisions:
                    print(i)
                    return (digest, [attempt, hash_dict[digest]])
            else:
                hash_dict[digest] = attempt

    
    def execute(self, iv, num):
        self._collisions = {}

        for _ in range(ceil(num / 2)):
            digest, inputs = self._find_collision(iv)
            key = (iv, digest)
            iv = digest
            
            if digest in self._collisions:
                self._collisions[key] = list(set(self._collisions[key] + inputs))
            else:
                self._collisions[key] = inputs
                
        return self._collisions


from samson.utilities.encoding import bytes_to_bitstring
from math import log, sqrt
from samson.stream_ciphers.rc4 import RC4
import operator
import json
import difflib
import scipy.special
import os

RC4_BIAS_MAP = [163, 0, 131, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 240, 17, 18, 0, 20, 21, 22, 0, 24, 25, 26, 0, 28, 29, 0, 31, 224, 33, 0, 0, 0, 0, 38, 0, 0, 0, 0, 0, 0, 0, 0, 0, 208, 0, 0, 0]


def longest_subsequence(seq_a: list, seq_b: list) -> list:
    """
    Finds the longest matching subsequence between two enumerable objects.

    Parameters:
        seq_a (list): First enumerable.
        seq_b (list): Second enumerable.
    
    Returns
        list: Longest subsequence.
    """
    seqMatch = difflib.SequenceMatcher(None, seq_a, seq_b)
    match = seqMatch.find_longest_match(0, len(seq_a), 0, len(seq_b))
    return seq_a[match.a: match.a + match.size]



def hamming_distance(bytes1: bytes, bytes2: bytes) -> int:
    """
    Calculates the Hamming distance between two byte-strings.

    Parameters:
        bytes1 (bytes): First byte-string.
        bytes2 (bytes): Second byte-string.
    
    Returns:
        int: Hamming distance.
    """
    assert len(bytes1) == len(bytes2)
    bitstring1 = bytes_to_bitstring(bytes1)
    bitstring2 = bytes_to_bitstring(bytes2)

    distance = 0
    for bit1, bit2 in zip(bitstring1, bitstring2):
        if bit1 != bit2: distance += 1

    return distance



def levenshtein_distance(seq_a: list, seq_b: list) -> int:
    """
    Calculates the Levenshtein Distance between two enumerable objects.

    Parameters:
        seq_a (list): First enumerable.
        seq_b (list): Second enumerable.
    
    Returns:
        int: Levenshtein Distance.
    """
    if len(seq_a) < len(seq_b):
        return levenshtein_distance(seq_b, seq_a)

    if len(seq_b) == 0:
        return len(seq_a)

    previous_row = range(len(seq_b) + 1)

    for i, c1 in enumerate(seq_a):
        current_row = [i + 1]

        for j, c2 in enumerate(seq_b):
            insertions = previous_row[j + 1] + 1 # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1       # than seq_b
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))

        previous_row = current_row
    
    return previous_row[-1]



def count_items(items: list) -> dict:
    """
    Counts the items in an enumerable object.

    Parameters:
        items (list): Enumerable of items.
    
    Returns:
        dict: Dictionary of {item, count}.
    """
    item_ctr = {curr_item: 0 for curr_item in items}

    for curr_item in items:
        item_ctr[curr_item] += 1

    return item_ctr



def chisquare(observed_dict: dict, expected_freq_dict: dict) -> float:
    """
    Calculates the Chi-squared score of an `observed_dict` against the `expected_freq_dict`.

    Parameters:
        observed_dict      (dict): Dictionary of observed items and their counts.
        expected_freq_dict (dict): Dictionary of expected items and their counts.
    
    Returns:
        float: Chi-squared score.
    """
    observed_len = sum([v for _k, v in observed_dict.items()])
    total = 0
    for key, freq_value in expected_freq_dict.items():
        if key in observed_dict:
            obs_val = observed_dict[key]
        else:
            obs_val = 0
        
        expected_number = observed_len * freq_value
        total += (expected_number - obs_val) ** 2 / expected_number

    
    for key, obs_value in observed_dict.items():
        if key not in expected_freq_dict:
            obs_val = obs_value
        else:
            obs_val = 0

        total += obs_val ** 2
    
    return total



def find_repeating_key_size(ciphertext: bytes, key_range: list) -> list:
    """
    Attempts to find the key size of a repeating XOR cipher.

    Parameters:
        ciphertext (bytes): Ciphertext to analyze.
        key_range   (list): List of key sizes to test.

    Returns:
        list: Sorted list of most likely key sizes.
    """
    key_distances = {}

    for size in key_range:
        size_sum = 0
        num_blocks = (len(ciphertext) // size)

        for block in range(num_blocks - 1):
            size_sum += hamming_distance(ciphertext[block * size: (block + 1) * size], ciphertext[(block + 1) * size: (block + 2) * size]) / size

        key_distances[size] = size_sum / num_blocks

    return sorted(key_distances.items(), key=operator.itemgetter(1))



def birthday_attack_analysis(bits: int, probability: float) -> float:
    """
    Determines the average number of attempts before a collision occurs against `bits` with `probability`.

    Parameters:
        bits          (int): Number of bits in the keyspace.
        probability (float): Target probability.
    
    Returns:
        float: Average number of attempts before collision.
    """
    return sqrt(2 * 2**bits * log(1/(1-probability)))



def num_expected_collisions(bits: int, num_inputs: int) -> float:
    """
    Calculates the number of expected collisions with `num_inputs` over `bits` keyspace.

    Parameters:
        bits       (int): Number of bits in the keyspace.
        num_inputs (int): Hypothetical number of inputs.
    
    Returns:
        float: Number of expected collisions.
    """
    return 2**(-bits)*scipy.special.comb(num_inputs, 2)



def generate_rc4_bias_map(ciphertexts):
    bias_map = [{} for i in range(256)]

    for c in ciphertexts:
        for i, byte in enumerate(c):
            if byte in bias_map[i]:
                bias_map[i][byte] = bias_map[i][byte] + 1
            else:
                bias_map[i][byte] = 1

    for i,_ in enumerate(bias_map):
        bias_map[i] = sorted(bias_map[i].items(), key=lambda kv: kv[1], reverse=True)

    return bias_map



def generate_random_rc4_bias_map(data=b'\x00' * 51, key_size=128, sample_size=2**20):
    ciphertexts = []

    for _ in range(sample_size):
        key = os.urandom(key_size // 8)
        cipher = RC4(key)
        ciphertexts.append(cipher.generate(len(data)) ^ data)


    return generate_rc4_bias_map(ciphertexts)



def incremental_rc4_bias_map_gen(filepath, start_idx=0, data=b'\x00' * 51, key_size=128, sample_size=2**30, chunk_size=2**24):
    if sample_size % chunk_size > 0:
        iteration_mod = 1
    else:
        iteration_mod = 0


    iterations = sample_size // chunk_size + iteration_mod

    for i in range(start_idx, iterations):
        if i == iterations - 1 and iteration_mod == 1:
            mod_sample_size = sample_size % chunk_size
        else:
            mod_sample_size = chunk_size
        
        bias_map = generate_random_rc4_bias_map(data, key_size, mod_sample_size)
        

        with open(filepath + ".{}".format(i), "w+") as f:
            f.write(json.dumps(bias_map))

        del bias_map



def merge_rc4_bias_map_files(base_path, num):
    bias_maps = []
    
    for i in range(num):
        with open("{}.{}".format(base_path, i)) as f:
            content = f.read()
        
        bias_maps.append(json.loads(content))

    return merge_rc4_bias_maps(bias_maps)



def merge_rc4_bias_maps(bias_maps):
    merged_map = [{} for i in range(256)]

    for bias_map in bias_maps:
        for i,_ in enumerate(bias_map):
            for k,v in bias_map[i]:
                if k in merged_map[i]:
                    merged_map[i][k] += v
                else:
                    merged_map[i][k] = v

    for i,_ in enumerate(merged_map):
        merged_map[i] = sorted(merged_map[i].items(), key=lambda kv: kv[1], reverse=True)

    return merged_map


from samson.auxiliary.markov_state import MarkovState
from samson.auxiliary.tokenizer_handler import TokenizerHandler

class MarkovChainHandler(TokenizerHandler):
    """
    Tokenize Handler that builds Markov chains.
    """

    def __init__(self):
        self.chain = MarkovState(count=1, probability=1, transitions={})
        self.curr_parsed_chain = self.chain


    def reset(self):
        self.curr_parsed_chain = self.chain

    
    def handle_token(self, token: str):
        """
        Handles a new token being found.

        Parameters:
            token (str): Found token.
        """
        if token in self.curr_parsed_chain.transitions:
            self.curr_parsed_chain.transitions[token].count += 1
        else:
            self.curr_parsed_chain.transitions[token] = MarkovState(count=1, probability=0, transitions={})

        self.curr_parsed_chain = self.curr_parsed_chain.transitions[token]



    def finalize(self) -> MarkovState:
        """
        Calculates the probabilities down the chains,

        Returns:
            MarkovState: The Markov chain.
        """
        self.chain.calculate_chain_probs()
        return self.chain


import numpy as np
import math
from copy import deepcopy
from types import FunctionType


class GreyWolfOptimizer:
  """
  Metaheuristic swarm-optmization algorithm.

>>> bRange = np.matrix([[0, 1],
... [-1, 1],
... [-1, 1],
... [-1, 1],
... [-1, 1]])

>>> gwo = GreyWolfOptimizer(200, 150, bRange, 0.01, oFunc)
>>> gwo.run()
  """

  def __init__(self, nWolves: int, nIters: int, boundsMat: np.ndarray, minConv: float, objFunc: FunctionType):
    """
    Parameters:
      nWolves             (int): Number of 'wolves' to spawn.
      nIters              (int): Number of iterations to run.
      boundsMat (numpy.ndarray): `p`x2 boundary matrix where `p` is the number of parameters.
      minConv           (float): Minminum convergence limit for early-out optimization.
      objFunc            (func): The objective function to *minimize*.
                                 NOTE: objFunc takes in wolf matrix
                                 and outputs cost vector of (nWolves, 1).
                                 This vector MUST be in the same order.
    """
    self.numWolves = nWolves
    self.maxIt = nIters
    self.wolves = np.zeros((0, boundsMat.shape[0] + 1))
    self.objFunc = objFunc
    self.boundaries = boundsMat
    self.bestWolves = np.zeros((0, boundsMat.shape[0] + 1))
    self.minConv = minConv

    numParams = boundsMat.shape[0]


    # Randomly initialize all wolves
    for w in range(self.numWolves):
      # Initialize wolf and add one to the wolf vector to hold cost
      wolf = np.zeros((1, numParams + 1))

      # Select random values within the boundaries
      for p in range(numParams):
        wolf[0][p] = np.random.uniform(self.boundaries[p, 0], self.boundaries[p, 1])

      self.wolves = np.concatenate((self.wolves, wolf))



  def run(self):
    """
    Runs the optimizer.

    Returns:
      int: Number of iterations ran. (NOTE: Solutions stored in `bestWolves`.)
    """
    numParams = self.wolves.shape[1] - 1
    min = np.transpose(self.boundaries[:, :1])
    max = np.transpose(self.boundaries[:, 1:2])

    last_top = 0.0
    max_conv_sample = 15
    curr_conv_sample = 0

    for i in range(self.maxIt):

      # Initialize variables
      a = 2 * (1 - (i ** 2 / self.maxIt ** 2))
      C = 2 * np.random.uniform(0, 1, (3 * self.numWolves, numParams))
      A = 2 * a * np.random.uniform(0, 1, (3 * self.numWolves, numParams)) - a

      # Calculate cost for each wolf
      self.wolves[:, numParams:] = self.objFunc(self.wolves)

      # Sort by best wolves using the cost column (last)
      self.wolves = np.array(self.wolves)
      self.wolves = self.wolves[self.wolves[:, numParams].argsort()]

      # Take top 3 wolves, don't take cost column
      topWolves = self.wolves[:3, :-1]
      topWolvesCost = self.wolves[:3]
      best_solution_cost = deepcopy(topWolvesCost[0, numParams:])
      self.bestWolves = np.concatenate((self.bestWolves, topWolvesCost[:1]))


      for w in range(self.numWolves):

        # Take every self.numWolves-th vector from matrices C and A
        currC = C[w::self.numWolves]
        currA = A[w::self.numWolves]

        # Repeat wolf w for each topWolf, so we can do a clean
        # subtraction (don't take cost column)
        currWolfMat = np.repeat(self.wolves[w, :-1].reshape(1, numParams), 3, axis=0)

        # Calculate Distance matrix
        D = np.absolute(np.multiply(topWolves, currC) - currWolfMat)

        currXs = topWolves - np.multiply(currA, D)

        # Take average as new solution
        mean = np.mean(currXs, axis=0)

        # Keep within boundaries
        boundsCheckedWolf = np.minimum(np.maximum(mean, min), max)
        self.wolves[w] = np.array(np.append(boundsCheckedWolf.tolist(), 0))

      self.wolves = np.concatenate((self.wolves[3:], topWolvesCost))

      # Determines whether we're making much progress
      # If not, we should break
      if last_top - best_solution_cost < self.minConv:
          curr_conv_sample += 1
      else:
          curr_conv_sample = 0

      if curr_conv_sample == max_conv_sample:
          return i

      last_top = best_solution_cost


    # Sort wolves
    self.bestWolves = self.bestWolves[self.bestWolves[:, numParams].argsort()]
    return i


# Simple test cost func
def oFunc(mat):
  return (mat[:, 0:1] * 3.0) + (mat[:, 1:2] * -2.0) + (mat[:, 2:3] * 0.75) + (mat[:, 3:4] * 10.0) + (mat[:, 4:5] * -0.8)


# More complex test case
def UF1(x):
  x = np.transpose(x)
  dim = x.shape[0]
  num = x.shape[1]
  tmp = np.zeros((dim, num))

  tiled = np.tile(x[0, :], dim - 1).reshape(dim - 1, num)

  temp1 = np.multiply(tiled, 6 * math.pi)
  temp2 = math.pi / dim * np.repeat(np.array(range(2, dim + 1)), num).reshape(dim - 1, num)


  tmp[1:dim, :] = (x[1:dim, :] - np.sin(temp1 + temp2)) ** 2

  evens = tmp[2:dim:2, :]
  odds = tmp[1:dim:2, :]

  if(evens.shape[0] > 1):
    tmp1 = np.sum(evens, axis=0)
  else:
    tmp1 = np.sum(evens)


  if(odds.shape[0] > 1):
    tmp2 = np.sum(odds, axis=0)
  else:
    tmp2 = np.sum(odds)


  y = np.zeros((2, num))

  y[0, :] = x[0, :] + 2.0 * tmp1 / len(range(2, dim, 2))
  y[1, :] = 1.0 - np.sqrt(x[0, :]) + 2.0 * tmp2 / len(range(1, dim, 2))
  return np.sum(y, axis=0).reshape(num, 1)


from samson.utilities.bytes import Bytes

class PKCS7(object):
    """
    PCKS#7 block cipher padding
    """

    def __init__(self, block_size: int=16):
        """
        Parameters:
            block_size (int): Block size of the cipher.
        """
        self.block_size = block_size
    

    def __repr__(self):
        return f"<PKCS7: block_size={self.block_size}>"

    def __str__(self):
        return self.__repr__()


    def pad(self, plaintext: bytes) -> Bytes:
        """
        Pads the plaintext.

        Parameters:
            plaintext (bytes): Plaintext to pad.
        
        Returns:
            Bytes: Padded plaintext.
        """
        padding = self.block_size - len(plaintext) % self.block_size
        return plaintext + Bytes(padding).stretch(padding)



    def unpad(self, plaintext: bytes, allow_padding_oracle: bool=False) -> Bytes:
        """
        Unpads the plaintext.

        Parameters:
            plaintext           (bytes): Plaintext to unpad.
            allow_padding_oracle (bool): Whether or not to explicitly create a padding oracle.
        
        Returns:
            Bytes: Unpadded plaintext.
        """
        last_block = plaintext[-self.block_size:]
        last_byte = last_block[-1]

        original_text, padding = plaintext[:len(plaintext) - last_byte], last_block[-last_byte:]
        if allow_padding_oracle and (len(padding) != last_byte or sum([last_byte != pad_char for pad_char in padding]) != 0):
            raise Exception('Invalid padding ;)')

        return original_text


from samson.utilities.manipulation import left_rotate, get_blocks
from samson.utilities.bytes import Bytes
import math
from copy import deepcopy

# https://en.wikipedia.org/wiki/Salsa20


def QUARTER_ROUND(a: int, b: int, c: int, d: int) -> (int, int, int, int):
    """
    Performs a quarter round of Salsa.

    Parameters:
        a (int): Salsa state variable.
        b (int): Salsa state variable.
        c (int): Salsa state variable.
        d (int): Salsa state variable.
    
    Returns:
        (int, int, int, int): New values for (a, b, c, d).
    """
    b = (b ^ left_rotate((a + d) & 0xFFFFFFFF,  7))
    c = (c ^ left_rotate((b + a) & 0xFFFFFFFF,  9))
    d = (d ^ left_rotate((c + b) & 0xFFFFFFFF, 13))
    a = (a ^ left_rotate((d + c) & 0xFFFFFFFF, 18))
    return a, b, c, d


class Salsa(object):
    """
    Salsa stream cipher

    Add-rotate-xor (ARX) structure.
    """

    def __init__(self, key: bytes, nonce: bytes, rounds: int=20, constant: bytes=b"expand 32-byte k"):
        """
        Parameters:
            key      (bytes): Key (128 or 256 bits).
            nonce    (bytes): Nonce (8 bytes).
            rounds     (int): Number of rounds to perform.
            constant (bytes): Constant used in generating the keystream (16 bytes).
        """
        self.key = key
        self.nonce = nonce
        self.rounds = rounds
        self.constant = constant



    def __repr__(self):
        return f"<Salsa: key={self.key}, nonce={self.nonce}, rounds={self.rounds}, constant={self.constant}>"

    def __str__(self):
        return self.__repr__()


    
    def full_round(self, block_num: int, state: list=None) -> Bytes:
        """
        Performs a full round of Salsa.

        Parameters:
            block_num (int): Current block number.
        
        Returns:
            Bytes: Keystream block.
        """
        ctr_bytes = int.to_bytes(block_num, 8, 'little')

        cons_blocks  = [int.from_bytes(block, 'little') for block in get_blocks(self.constant, 4)]
        key_blocks   = [int.from_bytes(block, 'little') for block in get_blocks(self.key, 4)]
        ctr_blocks   = [int.from_bytes(block, 'little') for block in get_blocks(ctr_bytes, 4)]
        nonce_blocks = [int.from_bytes(block, 'little') for block in get_blocks(self.nonce, 4)]

        x = state or [
            cons_blocks[0],  *key_blocks[:4],
            cons_blocks[1],  *nonce_blocks,
            *ctr_blocks,     cons_blocks[2],
            *key_blocks[4:], cons_blocks[3]
        ]

        x = deepcopy(x)
        tmp = deepcopy(x)

        for _ in range(self.rounds // 2):
            # Odd round
            x[ 0], x[ 4], x[ 8], x[12] = QUARTER_ROUND(x[ 0], x[ 4], x[ 8], x[12])
            x[ 5], x[ 9], x[13], x[ 1] = QUARTER_ROUND(x[ 5], x[ 9], x[13], x[ 1])
            x[10], x[14], x[ 2], x[ 6] = QUARTER_ROUND(x[10], x[14], x[ 2], x[ 6])
            x[15], x[ 3], x[ 7], x[11] = QUARTER_ROUND(x[15], x[ 3], x[ 7], x[11])

            # Even round
            x[ 0], x[ 1], x[ 2], x[ 3] = QUARTER_ROUND(x[ 0], x[ 1], x[ 2], x[ 3])
            x[ 5], x[ 6], x[ 7], x[ 4] = QUARTER_ROUND(x[ 5], x[ 6], x[ 7], x[ 4])
            x[10], x[11], x[ 8], x[ 9] = QUARTER_ROUND(x[10], x[11], x[ 8], x[ 9])
            x[15], x[12], x[13], x[14] = QUARTER_ROUND(x[15], x[12], x[13], x[14])

        for i in range(16):
            x[i] += tmp[i]

        return Bytes(b''.join([int.to_bytes(state_int & 0xFFFFFFFF, 4, 'little') for state_int in x]), byteorder='little')




    def yield_state(self, start_chunk: int=0, num_chunks: int=1, state: list=None):
        """
        Generates `num_chunks` chunks of keystream starting from `start_chunk`.

        Parameters:
            num_chunks  (int): Desired number of 64-byte keystream chunks.
            start_chunk (int): Chunk number to start at.
            state      (list): Custom state to be directly injected.
        
        Returns:
            generator: Keystream chunks.
        """
        for iteration in range(start_chunk, start_chunk + num_chunks):
            yield self.full_round(iteration)



    def generate(self, length: int) -> Bytes:
        """
        Generates `length` of keystream.

        Parameters:
            length (int): Desired length of keystream in bytes.
        
        Returns:
            Bytes: Keystream.
        """
        num_chunks = math.ceil(length / 64)

        return sum(list(self.yield_state(num_chunks=num_chunks)))[:length]


from samson.utilities.manipulation import get_blocks, transpose, stretch_key
from samson.utilities.general import rand_bytes
from samson.utilities.bytes import Bytes
import math
from collections import UserString

class Bitstring(UserString):
    def __init__(self, value, byteorder='big'):
        if type(value) is bytes or type(value) is bytearray or type(value) is Bytes:
            value = int.from_bytes(value, byteorder)
        
        if type(value) is int:
            value = bin(value)[2:]


        if not all([bit in ['0', '1'] for bit in str(value)]):
            raise Exception("Bitstrings can only contain 1's or 0's.")

        super().__init__(value)
        self.byteorder = byteorder


    def _format_return(self, value):
        return Bitstring(str(Bitstring(value, 'big').zfill(len(self))), self.byteorder)#.zfill(len(self))


    @staticmethod
    def wrap(str_like, byteorder='big'):
        if isinstance(str_like, Bitstring):
            return str_like
        else:
            return Bitstring(str_like, byteorder=byteorder)

    
    @staticmethod
    def random(size=16, byteorder='big'):
        return Bitstring(rand_bytes(math.ceil(size / 8)), byteorder=byteorder)[:size]


    def __repr__(self):
        return f'<Bitstring: {" ".join([str(chunk) for chunk in self.zfill(math.ceil(max(1, len(self)) / 8) * 8).chunk(8)])}, byteorder={self.byteorder}>'


    # Byte order will be the downfall of humanity.
    def _clean_byteorder(self, value):
        value.byteorder = 'big'
        return value.int()



    def __xor__(self, other):
        other = Bitstring.wrap(other, self.byteorder).bytes()
        result = self._clean_byteorder(self.bytes() ^ other)
        return self._format_return(result)


    def __rxor__(self, other):
        return self.__xor__(other)


    def __getitem__(self, index):
        return Bitstring(UserString.__getitem__(self, index), self.byteorder)
        

    def __and__(self, other):
        other = Bitstring.wrap(other, self.byteorder).bytes()
        result = self._clean_byteorder(self.bytes() & other)
        return self._format_return(result)


    def __rand__(self, other):
        return self.__and__(other)


    def __or__(self, other):
        other = Bitstring.wrap(other, self.byteorder).bytes()
        result = self._clean_byteorder(self.bytes() | other)
        return self._format_return(result)


    def __ror__(self, other):
        return self.__or__(other)


    def __add__(self, other):
        return Bitstring(UserString.__add__(self, other), self.byteorder)


    def __radd__(self, other):
        return Bitstring(UserString(other).__add__(self), self.byteorder)


    def __lshift__(self, num):
        return self._format_return(self.bytes() << num)


    def __rshift__(self, num):
        return self._format_return(self.bytes() >> num)


    def __invert__(self):
        max_val = 2 ** len(self) - 1
        return self._format_return(max_val - self.int())


    def lrot(self, amount, bits=None):
        return self._format_return(self.bytes().lrot(amount, bits=bits))



    def rrot(self, amount, bits=None):
        return self._format_return(self.bytes().rrot(amount, bits=bits))


    def chunk(self, size, allow_partials=False):
        return get_blocks(self, size, allow_partials)


    def transpose(self, size):
        return self._format_return(''.join(transpose(str(self), size)))


    def zfill(self, size):
        zeroes = max(size - len(self), 0) * '0'
        if self.byteorder == 'big':
            new_bitstring = zeroes + self
        else:
            new_bitstring = self + zeroes

        return Bitstring(new_bitstring, self.byteorder)



    def stretch(self, size, offset=0):
        return self._format_return(stretch_key(self, size, offset))

    
    def to_int(self):
        return self.bytes().int()


    def int(self):
        return self.to_int()

    
    def to_bytes(self):
        return Bytes([int(str(chunk), 2) for chunk in self.zfill(self.byte_length()).chunk(8, allow_partials=True)], self.byteorder)

    
    def bytes(self):
        return self.to_bytes()


    def byte_length(self):
        return math.ceil(len(self) / 8)


from types import FunctionType

class PaddingOracle(object):
    """
    Oracle that determines if a ciphertext has the correct padding or not. The `request` function must return a boolean indicating this.
    """

    def __init__(self, request_func: FunctionType):
        """
        Parameters:
            request_func (func): Function that takes in bytes and returns a boolean indicating whether the resulting plaintext has correct padding.
        """
        self.check_padding = request_func


# Ref: https://github.com/ajalt/python-sha1/blob/master/sha1.py
import struct
from samson.utilities.manipulation import left_rotate
from samson.utilities.bytes import Bytes
from samson.constructions.merkle_damgard_construction import MerkleDamgardConstruction

#h0, h1, h2, h3, h4
#_process_chunk
def compression_func(chunk, state):
    """Process a chunk of data and return the new digest variables."""
    assert len(chunk) == 64

    w = [0] * 80

    # Break chunk into sixteen 4-byte big-endian words w[i]
    for i in range(16):
        w[i] = struct.unpack(b'>I', chunk[i*4:i*4 + 4])[0]

    # Extend the sixteen 4-byte words into eighty 4-byte words
    for i in range(16, 80):
        w[i] = left_rotate(w[i-3] ^ w[i-8] ^ w[i-14] ^ w[i-16], 1)

    # Initialize hash value for this chunk
    h0, h1, h2, h3, h4 = bytes_to_state(state)

    a = h0
    b = h1
    c = h2
    d = h3
    e = h4

    for i in range(80):
        if 0 <= i <= 19:
            # Use alternative 1 for f from FIPS PB 180-1 to avoid bitwise not
            f = d ^ (b & (c ^ d))
            k = 0x5A827999
        elif 20 <= i <= 39:
            f = b ^ c ^ d
            k = 0x6ED9EBA1
        elif 40 <= i <= 59:
            f = (b & c) | (b & d) | (c & d)
            k = 0x8F1BBCDC
        elif 60 <= i <= 79:
            f = b ^ c ^ d
            k = 0xCA62C1D6

        a, b, c, d, e = ((left_rotate(a, 5) + f + e + k + w[i]) & 0xffffffff,
                        a, left_rotate(b, 30), c, d)

    # Add this chunk's hash to result so far
    h0 = (h0 + a) & 0xffffffff
    h1 = (h1 + b) & 0xffffffff
    h2 = (h2 + c) & 0xffffffff
    h3 = (h3 + d) & 0xffffffff
    h4 = (h4 + e) & 0xffffffff

    state = [h0, h1, h2, h3, h4]

    return Bytes(state_to_bytes(state))



def state_to_bytes(state):
    return int.to_bytes(sum(x<<(32*i) for i, x in enumerate(state[::-1])), 20, 'big')


def bytes_to_state(state_bytes):
    as_int = int.from_bytes(state_bytes, 'big')
    return [(as_int>>(32*i)) & 0xffffffff for i in range(4, -1, -1)]



class SHA1(MerkleDamgardConstruction):
    """
    Cryptographic hash function considered to be broken but is still widely used.
    """

    def __init__(self, initial_state: bytes=state_to_bytes([0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476, 0xC3D2E1F0])):
        """
        Parameters:
            initial_state (bytes): (Optional) Initial internal state.
        """
        if type(initial_state) is list:
            initial_state = state_to_bytes(initial_state)

        super().__init__(
            initial_state=initial_state,
            compression_func=compression_func,
            digest_size=20,
        )


    def __repr__(self):
        return f"<SHA1: initial_state={self.initial_state}, block_size={self.block_size}>"


    def __str__(self):
        return self.__repr__()


class TokenizerHandler(object):
    """
    Base class for TokenizerHandlers.
    """

    def __init__(self):
        pass


    def reset(self):
        """Resets the internal for the next sample."""
        raise NotImplementedError()

    
    def handle_token(self, token: str):
        """Handles a new token being found."""
        raise NotImplementedError()


    def finalize(self) -> object:
        """Makes final modifications to state and returns result."""
        raise NotImplementedError()


from samson.utilities.bytes import Bytes

# https://en.wikipedia.org/wiki/HMAC
class HMAC(object):
    """
    Hash-based message authentication code using a generic interface to hash functions.
    """

    def __init__(self, key: bytes, hash_obj: object):
        """
        Parameters:
            key       (bytes): Bytes-like object to key the HMAC.
            hash_obj (object): Instantiated object with compatible hash interface.
        """
        self.key = Bytes.wrap(key)
        self.hash_obj = hash_obj

        key_prime = self.key
        if len(self.key) > self.hash_obj.block_size:
            key_prime = self.hash_obj.hash(self.key)

        self.key_prime = key_prime + b'\x00' * (self.hash_obj.block_size - len(key_prime))
        self.outer_key_pad = self.key_prime ^ Bytes(b'\x5c').stretch(self.hash_obj.block_size)
        self.inner_key_pad = self.key_prime ^ Bytes(b'\x36').stretch(self.hash_obj.block_size)


    def __repr__(self):
        return f"<HMAC: key={self.key}, key_prime={self.key_prime}, outer_key_pad={self.outer_key_pad}, inner_key_pad={self.inner_key_pad}>"

    def __str__(self):
        return self.__repr__()


    def generate(self, message: bytes) -> Bytes:
        """
        Generates a keyed MAC for `message`.

        Parameters:
            message (bytes): Message to generate a MAC for.
        
        Returns:
            Bytes: The MAC.
        """
        return self.hash_obj.hash(self.outer_key_pad + self.hash_obj.hash(self.inner_key_pad + message))


from copy import deepcopy
from samson.utilities.general import rand_bytes
from sympy.matrices import Matrix, GramSchmidt
from sympy import isprime, Poly
from sympy.abc import x
import math

def gcd(a: int, b: int) -> int:
    """
    Recursively computes the greatest common denominator.

    Parameters:
        a (int): First integer.
        b (int): Second integer.
    
    Returns:
        int: GCD of `a` and `b`.
    """
    if b == 0:
        return a
    else:
        return gcd(b, a % b)


# https://anh.cs.luc.edu/331/notes/xgcd.pdf
def xgcd(a: int, b: int) -> (int, int, int):
    """
    Extended Euclidean algorithm form of GCD.
    `ax + by = gcd(a, b)`

    Parameters:
        a (int): First integer.
        b (int): Second integer.
    
    Returns:
        (int, int, int): Formatted as (GCD, x, y).
    """
    prevx, x = 1, 0; prevy, y = 0, 1
    while b:
        q = a//b
        x, prevx = prevx - q*x, x
        y, prevy = prevy - q*y, y
        a, b = b, a % b
    return a, prevx, prevy



def lcm(a: int, b: int) -> int:
    """
    Calculates the least common multiple of `a` and `b`.

    Parameters:
        a (int): First integer.
        b (int): Second integer.
    
    Returns:
        int: Least common multiple.
    """
    return a // gcd(a, b) * b



def mod_inv(a: int, n: int) -> int:
    """
    Calculates the modular inverse according to
    https://en.wikipedia.org/wiki/Euclidean_algorithm#Linear_Diophantine_equations
    and https://en.wikipedia.org/wiki/Extended_Euclidean_algorithm.

    Parameters:
        a (int): Integer.
        n (int): Modulus.
    
    Returns:
        int: Modular inverse of `a` over `n`.
    """

    _, x, _ = xgcd(a, n)

    if (a * x) % n != 1:
        raise Exception("'a' is not invertible")
    
    if x < 0:
        x = x + n

    return x
    

def modexp (g: int, u: int, p: int) -> int:
    """
    Computes `s = (g ^ u) mod p` (see Bruce Schneier's book, _Applied Cryptography_ p. 244).

    Parameters:
        g (int): Base.
        u (int): Exponent.
        p (int): Modulus.
    
    Returns:
        int: Modular exponentiation.
    """
    s = 1
    while u != 0:
        if u & 1:
            s = (s * g)%p
        u >>= 1
        g = (g * g)%p
    return s


# https://stackoverflow.com/questions/23621833/is-cube-root-integer
def kth_root(n: int, k: int) -> int:
    """
    Calculates the `k`-th integer root of `n`.

    Parameters:
        n (int): Integer.
        k (int): Root (e.g. 2).
    
    Returns:
        int: `k`-th integer root of `n
    """
    lb,ub = 0,n #lower bound, upper bound
    while lb < ub:
        guess = (lb+ub)//2
        if pow(guess,k) < n: lb = guess+1
        else: ub = guess
    return lb


def crt(residues: list, moduli: list) -> (int, int):
    """
    Performs the Chinese Remainder Theorem and returns the computed `x` and modulus.

    Parameters:
        residues (list): Residues of `x` in order relative to `moduli`.
        moduli   (list): Moduli of the residues.
    
    Returns:
        (int, int): Formatted as (computed `x`, modulus).
    """
    assert len(residues) == len(moduli)
    x = residues[0]
    Nx = moduli[0]

    for i in range(1, len(residues)):
        x = (mod_inv(Nx, moduli[i]) * (residues[i] - x)) * Nx + x
        Nx = Nx * moduli[i]

    return x % Nx, Nx



def legendre(a: int, p: int) -> int:
    """
    Calculates the Legendre symbol of `a` mod `p`. Nonzero quadratic residues mod `p` return 1 and nonzero, non-quadratic residues return -1. Zero returns 0.

    Parameters:
        a (int): Integer.
        p (int): Modulus.
    
    Returns:
        int: Legendre symbol.
    """
    return pow(a, (p - 1) // 2, p)


# https://crypto.stackexchange.com/questions/22919/explanation-of-each-of-the-parameters-used-in-ecc
# https://www.geeksforgeeks.org/find-square-root-modulo-p-set-2-shanks-tonelli-algorithm/
# https://rosettacode.org/wiki/Tonelli-Shanks_algorithm#Python


def tonelli(n: int, p: int) -> int:
    """
    Performs the Tonelli-Shanks algorithm for calculating the square root of `n` mod `p`.

    Parameters:
        n (int): Integer.
        p (int): Modulus.
    
    Returns:
        int: Square root of `n` mod `p`.
    """
    assert legendre(n, p) == 1, "not a square (mod p)"
    q = p - 1
    s = 0
    while q % 2 == 0:
        q //= 2
        s += 1
    if s == 1:
        return pow(n, (p + 1) // 4, p)
    for z in range(2, p):
        if p - 1 == legendre(z, p):
            break
    c = pow(z, q, p)
    r = pow(n, (q + 1) // 2, p)
    t = pow(n, q, p)
    m = s
    t2 = 0
    while (t - 1) % p != 0:
        t2 = (t * t) % p
        for i in range(1, m):
            if (t2 - 1) % p == 0:
                break
            t2 = (t2 * t2) % p
        b = pow(c, 1 << (m - i - 1), p)
        r = (r * b) % p
        c = (b * b) % p
        t = (t * c) % p
        m = i
    return r



# https://github.com/orisano/olll/blob/master/olll.py
# https://en.wikipedia.org/wiki/Lenstra%E2%80%93Lenstra%E2%80%93Lov%C3%A1sz_lattice_basis_reduction_algorithm
def lll(in_basis: list, delta=0.75) -> Matrix:
    """
    Performs the Lenstra–Lenstra–Lovász lattice basis reduction algorithm.

    Parameters:
        in_basis (list): List of Matrix objects representing the original basis.
        delta   (float): Minimum optimality of the reduced basis.
    
    Returns:
        Matrix: Reduced basis.
    """
    basis = deepcopy(in_basis)
    n = len(basis)
    ortho = GramSchmidt(basis)

    def mu(i, j):
        return ortho[j].dot(basis[i]) / ortho[j].dot(ortho[j])

    k = 1
    while k < n:
        for j in range(k - 1, -1, -1):
            mu_kj = mu(k, j)
            if abs(mu_kj) > 0.5:
                basis[k] = basis[k] - basis[j] * round(mu_kj)
                ortho = GramSchmidt(basis)


        if ortho[k].dot(ortho[k]) >= (delta - mu(k, k - 1)**2) * (ortho[k - 1].dot(ortho[k - 1])):
            k += 1
        else:
            basis[k], basis[k - 1] = deepcopy(basis[k - 1]), deepcopy(basis[k])
            ortho = GramSchmidt(basis)
            k = max(k - 1, 1)

    return Matrix([list(map(int, b)) for b in basis])



def generate_superincreasing_seq(length: int, max_diff: int, starting: int=0) -> list:
    """
    Generates a superincreasing sequence.

    Parameters:
        length   (int): Number of elements to generate.
        max_diff (int): Maximum difference between the sum of all elements before and the next element.
        starting (int): Minimum starting integer.
    
    Returns:
        list: List of the superincreasing sequence.
    """
    seq = []

    last_sum = starting
    for _ in range(length):
        delta = int.from_bytes(rand_bytes(math.ceil(math.log(max_diff, 256))), 'big') % max_diff
        seq.append(last_sum + delta)
        last_sum = sum(seq)

    return seq



def find_coprime(p: int, search_range: list) -> int:
    """
    Attempts to find an integer coprime to `p`.

    Parameters:
        p             (int): Integer to find coprime for.
        search_range (list): Range to look in.
    
    Returns:
        int: Integer coprime to `p`.
    """
    for i in search_range:
        if gcd(p, i) == 1:
            return i



def find_prime(bits: int) -> int:
    """
    Finds a prime of `bits` bits.

    Parameters:
        bits (int): Bit length of prime.
    
    Returns:
        int: Prime.
    """
    rand_num = int.from_bytes(rand_bytes(math.ceil(bits / 8)), 'big')
    rand_num |= 2**(bits - 1)
    return next_prime(rand_num)



def next_prime(start_int: int) -> int:
    """
    Finds the next prime.

    Parameters:
        start_int (int): Integer to start search at.
    
    Returns:
        int: Prime.
    """
    start_int |= 1
    while not isprime(start_int):
        start_int += 2
    
    return start_int



# https://en.wikipedia.org/wiki/Berlekamp%E2%80%93Massey_algorithm
def berlekamp_massey(output_list: list) -> Poly:
    """
    Performs the Berlekamp-Massey algorithm to find the shortest LFSR for a binary output sequence.

    Parameters:
        output_list (list): Output of LFSR.
    
    Returns:
        Poly: Polyomial that represents the shortest LFSR.
    """
    n = len(output_list)
    b = [1] + [0] * (n - 1)
    c = [1] + [0] * (n - 1)

    L = 0
    m = -1

    i  = 0
    while i < n:
        out_vec = output_list[i - L:i][::-1]
        c_vec = c[1:L+i]
        d = output_list[i] + sum([s_x * c_x for s_x, c_x in zip(out_vec, c_vec)]) % 2

        if d == 1:
            t = deepcopy(c)
            p = [0] * n
            for j in range(L):
                if b[j] == 1:
                    p[j + i - m] = 1
            
            c = [(c_x + p_x) % 2 for c_x, p_x in zip(c, p)]

            if L <= i / 2:
                L = i + 1 - L
                m = i
                b = t
        
        i += 1

    return Poly(c[:L + 1], x)


def is_power_of_two(n: int) -> bool:
    """
    Determines if `n` is a power of two.

    Parameters:
        n (int): Integer.
    
    Returns:
        bool: Whether or not `n` is a power of two.
    """
    return n != 0 and (n & (n - 1) == 0)



class Analyzer(object):
    """
    Base class for all Analyzers.

    Analyzers take in a byte-string and outputs a score where high scores match a context better.
    The score does not have to be absolute.
    """

    def __init__(self):
        pass


    def analyze(self, in_bytes: bytes) -> float:
        """
        Analyzes a bytes-like object and returns its "score".

        Parameters:
            in_bytes (bytes): The bytes-like object to be analyzed.

        Returns:
            float: The score.
        """
        raise NotImplementedError("The `analyze` method must be implemented by a subclass.")


    def select_highest_scores(self, in_list: list, num: int=1) -> list:
        """
        Analyzes a list `in_list`, sorts the list, and returns the top `num` scores.

        Parameters:
            in_list (list): The list of byte-like objects to be analyzed.
            num      (int): Number of results to return.

        Returns:
            list: `in_list` sorted and truncated.
        """
        return sorted(in_list, key=lambda item: self.analyze(item), reverse=True)[:num]


from samson.utilities.math import mod_inv
from samson.utilities.bytes import Bytes

class DSA(object):
    """
    Digital Signature Algorithm
    """

    def __init__(self, hash_obj: object, p: int=None, q: int=None, g: int=None, x: int=None):
        """
        Parameters:
            hash_obj (object): Instantiated object with compatible hash interface.
            p           (int): (Optional) Prime modulus.
            q           (int): (Optional) Prime modulus.
            g           (int): (Optional) Generator.
            x           (int): (Optional) Private key.
        """
        self.p = p or 0x800000000000000089E1855218A0E7DAC38136FFAFA72EDA7859F2171E25E65EAC698C1702578B07DC2A1076DA241C76C62D374D8389EA5AEFFD3226A0530CC565F3BF6B50929139EBEAC04F48C3C84AFB796D61E5A4F9A8FDA812AB59494232C7D2B4DEB50AA18EE9E132BFA85AC4374D7F9091ABC3D015EFC871A584471BB1
        self.q = q or 0xF4F47F05794B256174BBA6E9B396A7707E563C5B
        self.g = g or 0x5958C9D3898B224B12672C0B98E06C60DF923CB8BC999D119458FEF538B8FA4046C8DB53039DB620C094C9FA077EF389B5322A559946A71903F990F1F7E0E025E2D7F7CF494AFF1A0470F5B64C36B625A097F1651FE775323556FE00B3608C887892878480E99041BE601A62166CA6894BDD41A7054EC89F756BA9FC95302291
        
        self.x = x or Bytes.random((self.q.bit_length() + 7) // 8).int() % self.q
        self.y = pow(self.g, self.x, self.p)
        self.hash_obj = hash_obj

    
    def __repr__(self):
        return f"<DSA: hash_obj={self.hash_obj}, p={self.p}, q={self.q}, g={self.g}, x={self.x}, y={self.y}>"


    def __str__(self):
        return self.__repr__()
    
    
    def sign(self, message: bytes, k: int=None) -> (int, int):
        """
        Signs a `message`.

        Parameters:
            message (bytes): Message to sign.
            k         (int): (Optional) Ephemeral key.
        
        Returns:
            (int, int): Signature formatted as (r, s).
        """
        k = k or max(1, Bytes.random((self.q.bit_length() + 7) // 8).int() % self.q)
        inv_k = mod_inv(k, self.q)
        r = pow(self.g, k, self.p) % self.q
        s = (inv_k * (self.hash_obj.hash(message) + self.x * r)) % self.q
        return (r, s)
    
    
    def verify(self, message: bytes, sig: (int, int)) -> bool:
        """
        Verifies a `message` against a `sig`.

        Parameters:
            message  (bytes): Message.
            sig ((int, int)): Signature of `message`.
        
        Returns:
            bool: Whether the signature is valid or not.
        """
        (r, s) = sig
        w = mod_inv(s, self.q)
        u_1 = (self.hash_obj.hash(message) * w) % self.q
        u_2 = (r * w) % self.q
        v = (pow(self.g, u_1, self.p) * pow(self.y, u_2, self.p) % self.p) % self.q
        return v == r

    
    # Confirmed works on ECDSA as well
    def derive_k_from_sigs(self, msg_a: bytes, sig_a: (int, int), msg_b: bytes, sig_b: (int, int)) -> int:
        """
        Derives `k` from signatures that share an `r` value.

        Parameters:
            msg_a      (bytes): Message A.
            msg_b      (bytes): Message B.
            sig_a ((int, int)): Signature of `msg_a`.
            sig_b ((int, int)): Signature of `msg_b`.

        Returns:
            int: Derived `k`.
        """
        (r_a, s_a) = sig_a
        (r_b, s_b) = sig_b
        assert r_a == r_b
        
        s = (s_a - s_b) % self.q
        m = (self.hash_obj.hash(msg_a) - self.hash_obj.hash(msg_b)) % self.q
        return mod_inv(s, self.q) * m % self.q
    
    
    # Confirmed works on ECDSA as well
    def derive_x_from_k(self, message: bytes, k: int, sig: (int, int)) -> int:
        """
        Derives `x` from a known `k`.

        Parameters:
            message  (bytes): Message.
            k          (int): `k` used in `message`'s signature.
            sig ((int, int)): Signature of `message`.
        
        Returns:
            int: Derived `x`.
        """
        (r, s) = sig
        return ((s * k) - self.hash_obj.hash(message)) * mod_inv(r, self.q) % self.q


from samson.hashes.md5 import MD5
import tempfile
import subprocess

import logging
log = logging.getLogger(__name__)



class HashClashCollider(object):
    """
    Finds MD5 collisions with HashClash. Used in the Nostradamus attack.
    """

    def __init__(self, hash_clash_base_location: str):
        """
        Parameters:
            hash_clash_base_location (str): Path to HashClash's location on disk.
        """
        self.hash_clash_base_location = hash_clash_base_location
        self.hasher = MD5()
        self.hasher.pad_func = lambda x: x


    
    def find_collision(self, p1: bytes, p2: bytes) -> (bytes, bytes, bytes):
        """
        Finds a collision using HashClash.

        Parameters:
            p1 (bytes): First sample.
            p2 (bytes): Second sample.
        
        Returns:
            (bytes, bytes, bytes): Tuple of bytes representing the collision as (p1_suffix, p2_suffix, self.hasher.hash(p1 + p1_suffix)).
        """
        tmp_dir = tempfile.TemporaryDirectory()

        log.info('Creating new directory: {}'.format(tmp_dir))

        p1_tmp = tempfile.NamedTemporaryFile(dir=tmp_dir.name)
        p2_tmp = tempfile.NamedTemporaryFile(dir=tmp_dir.name)

        # Write chosen-prefixes to temp files
        with open(p1_tmp, 'wb') as f:
            f.write(p1)

        with open(p2_tmp, 'wb') as f:
            f.write(p1)

        command = 'cd {} && {}/scripts/cpc.sh {} {}'.format(tmp_dir.name, self.hash_clash_base_location, p1_tmp.name, p2_tmp.name)
        log.info('Executing: {}'.format(command))
        process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)

        for line in iter(process.stdout.readline, b''):
            log.debug(line)


        # Read crafted values from files
        with open(p1_tmp + '.coll', 'rb') as f:
            p1_suffix = f.read()[len(p1):]

        with open(p2_tmp + '.coll', 'rb') as f:
            p2_suffix = f.read()[len(p2):]


        return p1_suffix, p2_suffix, self.hasher.hash(p1 + p1_suffix)


# Takes in a list of tokens
# Generates a Markov Chain
# Must work for delimited and non-delimited inputs
# Must grab the longest string possible
from samson.auxiliary.token_list_handler import TokenListHandler
from samson.auxiliary.tokenizer_handler import TokenizerHandler

class Tokenizer(object):
    """
    Splits a string into the longest 'tokens' parameterized bt the `token_list`. Works for delimited
    and non-delimited inputs.
    """

    def __init__(self, token_list: list, token_handler: TokenizerHandler=TokenListHandler, delimiter: str=' '):
        """
        Parameters:
            token_list                (list): List of possible tokens, e.g. wordlist.
            token_handler (TokenizerHandler): Instantiable class.
            delimiter                  (str): (Optional) Delimiter to split samples apart.
        """
        self.token_list = token_list
        self.token_handler = token_handler
        self.delimiter = delimiter

        self.processed_tokens = {}

        for token in self.token_list:
            curr_dict = self.processed_tokens

            for char in token:
                if char not in curr_dict:
                    curr_dict[char] = {}
                curr_dict = curr_dict[char]

            # '$$' is the flag for end of string
            curr_dict['$$'] = None



    def tokenize(self, samples: list) -> object:
        """
        Tokenizes a list of samples.

        Parameters:
            samples (list): List of string samples to break into tokens.
        
        Returns:
            object: The finalized return of the `token_handler`.
        """
        ################
        #    STEP 1    #
        # Create chain #
        ################
        token_handler = self.token_handler()

        for sample in samples:
            parts = sample.split(self.delimiter)
            token_handler.reset()

            for part in parts:
                token = ''
                last_largest_substring = ''
                curr_token_chain = self.processed_tokens

                # We'll use a while loop, so we have more control over the counter
                char_counter = 0

                while char_counter < len(part):
                    char = part[char_counter]

                    if char in curr_token_chain:
                        token += char
                        curr_token_chain = curr_token_chain[char]

                        # This is a valid substring, and there's only an end-of-token flag left
                        # OR
                        # This is a valid substring, and we've reached the end of the part
                        if curr_token_chain == {'$$': None} or ('$$' in curr_token_chain and char_counter + 1 == len(part)):
                            token_handler.handle_token(token)

                            token = ''
                            last_largest_substring = ''
                            curr_token_chain = self.processed_tokens

                        # This is a valid substring, but it may not be the whole string.
                        # We'll keep track of it in case it is.
                        elif '$$' in curr_token_chain:
                            last_largest_substring = token


                    # It's not a valid token; we'll need to rollback.
                    # Example:

                    # tokenizer = Tokenizer(['abc', 'hello', 'adam', 'hiya'])
                    # chain = tokenizer.generate_chain(['adabcadam'])
                    # For 'adabcadam', we'll get to 'ada', but 'b' will not be a valid state.
                    # We then must rollback to 'd'.
                    else:
                        # The entire thing isn't a valid token, but a substring is.
                        substring_mod = 0
                        if last_largest_substring != '':
                            token_handler.handle_token(last_largest_substring)
                            substring_mod = len(last_largest_substring) - 1
                            

                        char_counter -= len(token) - substring_mod
                        token = ''
                        last_largest_substring = ''
                        curr_token_chain = self.processed_tokens

                    char_counter += 1


        return token_handler.finalize()


from samson.constructions.merkle_damgard_construction import MerkleDamgardConstruction
from samson.utilities.manipulation import left_rotate
from samson.utilities.bytes import Bytes
import math

# https://rosettacode.org/wiki/MD5/Implementation#Python

rotate_amounts = [7, 12, 17, 22, 7, 12, 17, 22, 7, 12, 17, 22, 7, 12, 17, 22,
                  5,  9, 14, 20, 5,  9, 14, 20, 5,  9, 14, 20, 5,  9, 14, 20,
                  4, 11, 16, 23, 4, 11, 16, 23, 4, 11, 16, 23, 4, 11, 16, 23,
                  6, 10, 15, 21, 6, 10, 15, 21, 6, 10, 15, 21, 6, 10, 15, 21]
 
constants = [int(abs(math.sin(i+1)) * 2**32) & 0xFFFFFFFF for i in range(64)]
 
functions = 16*[lambda b, c, d: (b & c) | (~b & d)] + \
            16*[lambda b, c, d: (d & b) | (~d & c)] + \
            16*[lambda b, c, d: b ^ c ^ d] + \
            16*[lambda b, c, d: c ^ (b | ~d)]
 
index_functions = 16*[lambda i: i] + \
                  16*[lambda i: (5*i + 1)%16] + \
                  16*[lambda i: (3*i + 5)%16] + \
                  16*[lambda i: (7*i)%16]
 




def state_to_bytes(state):
    return int.to_bytes(sum(x<<(32*i) for i, x in enumerate(state)), 16, 'little')


def bytes_to_state(state_bytes):
    as_int = int.from_bytes(state_bytes, 'little')
    return [(as_int>>(32*i)) & 0xffffffff for i in range(4)]



def compression_func(message, state):
    new_state = bytes_to_state(state)

    for chunk_ofst in range(0, len(message), 64):
        a, b, c, d = new_state
        chunk = message[chunk_ofst:chunk_ofst+64]

        for i in range(64):
            f = functions[i](b, c, d)
            g = index_functions[i](i)
            to_rotate = a + f + constants[i] + int.from_bytes(chunk[4*g:4*g+4], byteorder='little')
            new_b = (b + left_rotate(to_rotate, rotate_amounts[i])) & 0xFFFFFFFF
            a, b, c, d = d, new_b, b, c

        for i, val in enumerate([a, b, c, d]):
            new_state[i] += val
            new_state[i] &= 0xFFFFFFFF

    return Bytes(state_to_bytes(new_state))



class MD5(MerkleDamgardConstruction):
    """
    Popular but completely broken cryptographic hash function.
    """

    def __init__(self, initial_state: bytes=state_to_bytes([0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476])):
        """
        Parameters:
            initial_state (bytes): (Optional) Initial internal state.
        """
        super().__init__(
            initial_state=initial_state,
            compression_func=compression_func,
            digest_size=16,
            endianness='little'
        )


    def __repr__(self):
        return f"<MD5: initial_state={self.initial_state}, block_size={self.block_size}>"


    def __str__(self):
        return self.__repr__()


from samson.utilities.bytes import Bytes
from samson.utilities.math import mod_inv
from samson.hashes.md5 import MD5
from samson.protocols.diffie_hellman import DiffieHellman
from types import FunctionType

# https://asecuritysite.com/encryption/dragonfly
# https://tools.ietf.org/html/draft-harkins-tls-dragonfly-04
class Dragonfly(object):
    """
    Dragonfly zero-knowledge proof.

    Used in WPA3.
    """

    def __init__(self, key: bytes, H: FunctionType=lambda m: MD5().hash(m), q: int=DiffieHellman.MODP_2048):
        """
        Parameters:
            key (bytes): Bytes-like object shared by both parties to authenticate each other.
            H    (func): Cryptographic hash function. Takes in bytes and returns the hash digest.
            q     (int): Modulus.
        """
        self.key = key
        self.q = q
        self.A = Bytes.random(16).to_int()
        self.a = Bytes.random(16).to_int()
        self.H = H

    
    def __repr__(self):
        return f"<Dragonfly: key={self.key}, H={self.H}, A={self.A}, a={self.a}, q={self.q}>"

    def __str__(self):
        return self.__repr__()


    def get_challenge(self) -> int:
        """
        Gets the challenge.

        Returns:
            int: The challenge.
        """
        sA = self.a + self.A

        PE = int.from_bytes(self.H(self.key)[:8], 'big')
        eA = mod_inv(pow(PE, self.A, self.q), self.q)
        return pow(PE, sA, self.q), eA


    def derive_key(self, challenge: int) -> int:
        """
        Derives the shared key from the other instance's challenge.

        Parameters:
            challenge (int): The other instance's challenge.
        
        Returns:
            int: Shared key.
        """
        PEsA, eA = challenge
        return pow(PEsA * eA, self.a, self.q)


from samson.hashes.keccak import Keccak

class SHA3(object):
    """
    Contains various SHA3 functions.
    """

    @staticmethod
    def build_sha3(r: int, c: int, bits: int, padding: int) -> Keccak:
        """
        Parameters:
            r       (int): The block size of the Keccak function.
            c       (int): The capacity of the Keccak function.
            bits    (int): The bits of output for the Keccak function.
            padding (int): The SHA3, domain-specific padding number.
        """
        k = Keccak(r, c, bits)

        def pad(in_bytes):
            bit_rate_bytes = (k.r + 7) // 8
            pad_len = bit_rate_bytes - (len(in_bytes) % bit_rate_bytes)

            if pad_len == 0:
                pad_len = bit_rate_bytes

            if pad_len == 1:
                return in_bytes + bytes([padding + 0x80])
            else:
                return in_bytes + bytes([padding] + ([0] * (pad_len - 2)) + [0x80])

        
        k.pad_func = pad
        return k


    @staticmethod
    def K224() -> Keccak:
        return SHA3.build_sha3(1152, 448, 224, 0x06)


    @staticmethod
    def K256() -> Keccak:
        return SHA3.build_sha3(1088, 512, 256, 0x06)


    @staticmethod
    def K384() -> Keccak:
        return SHA3.build_sha3(832, 768, 384, 0x06)


    @staticmethod
    def K512() -> Keccak:
        return SHA3.build_sha3(576, 1024, 512, 0x06)


    @staticmethod
    def SHAKE128(bit_length: int) -> Keccak:
        return SHA3.build_sha3(1344, 256, bit_length, 0x1F)


    @staticmethod
    def SHAKE256(bit_length: int) -> Keccak:
        return SHA3.build_sha3(1088, 512, bit_length, 0x1F)


from samson.utilities.bytes import Bytes
from samson.utilities.math import tonelli, mod_inv
from fastecdsa.point import Point
from fastecdsa.curve import Curve
import random

class DualEC(object):
    """
    Implementation of the NSA's backdoored DRBG.
    """

    def __init__(self, P: Point, Q: Point, seed: int):
        """
        Parameters:
            P  (Point): Elliptical curve point `P`.
            Q  (Point): Elliptical curve point `Q`.
            seed (int): Initial value.
        """
        self.P = P
        self.Q = Q
        self.t = seed
        self.r = None


    def __repr__(self):
        return f"<DualEC: P={self.P}, Q={self.Q}, t={self.t}>"


    def __str__(self):
        return self.__repr__()


    def generate(self) -> Bytes:
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        s = (self.t * self.P).x
        self.t = s
        self.r = (s * self.Q).x

        return Bytes(int.to_bytes(self.r, 32, 'big')).zfill(32)[2:]


    @staticmethod
    def generate_backdoor(curve: Curve) -> (Point, Point, int):
        """
        Generates backdoored parameters.

        Parameters:
            curve (Curve): `fastecdsa` Curve to use.
        
        Returns:
            (Point Point, int): Result formatted as (P, backdoored Q, backdoor d)
        """
        P = curve.G
        d = random.randint(2, curve.q)
        e = mod_inv(d, curve.q)
        Q = e * P

        return P, Q, d


    @staticmethod
    def derive_from_backdoor(P: Point, Q: Point, d: int, observed_out: bytes) -> list:
        """
        Recovers the internal state of a Dual EC generator and builds a replica.

        Parameters:
            P            (Point): Elliptical curve point `P`.
            Q            (Point): Elliptical curve point `Q`.
            d              (int): Backdoor that relates Q to P.
            observed_out (bytes): Observed output from the compromised Dual EC generator.
        
        Returns:
            list: List of possible internal states.
        """

        assert len(observed_out) >= 30

        curve = P.curve

        r1 = observed_out[:30]
        r2 = observed_out[30:]

        possible_states = []

        for i in range(2**16):
            test_r1 = int.to_bytes(i, 2, 'big') + r1
            test_x = int.from_bytes(test_r1, 'big')
            try:
                y_2 = test_x ** 3 + (curve.a * test_x) + curve.b
                y = tonelli(y_2, curve.p)
                R = Point(test_x, y)
                dR = d * R
                test_r2 = dR.x * Q

                if int.to_bytes(test_r2.x, 32, 'big')[2:2 + len(r2)] == r2:
                    possible_states.append(DualEC(P, Q, dR.x))
            except Exception as _:
                pass

        return possible_states


from copy import deepcopy
from types import FunctionType
from samson.utilities.bytes import Bytes

import logging
log = logging.getLogger(__name__)


class NostradamusAttack(object):
    """
    Performs an precomputed multicollision attack.

    The Nostradamus attack is an extension to an iterated-hash multicollision. By precomputing a binary collision tree,
    you can coerce several initial values into a singular final collision. This was used to "predict" the 2008 presidential
    elections. It's highly recommended to use an efficient collision function like an implementation of Wang's attack (Google "Hashclash").
    """

    # 'k' is the number of levels the tree will have
    # `construction_func` takes in a message and POSSIBLY an IV and outputs a generator that yields the intermediary state


    def __init__(self, k: int, collision_func: FunctionType, output_size: int, prefixes: list=None):
        """
        Parameters:
            k               (int): Number of levels the tree will have.
            collision_func (func): Function that finds a collision in the hash function. Should return a tuple
                                   (input_bytes1, input_bytes2, colliding_state).
            output_size     (int): Size of the hash output.
            prefixes       (list): List of bytes-like prefixes. These are the starting nodes of the tree.
                                   You should set this to values you want to guarantee are in the tree.
        """
        self.k = k
        self.prefixes = prefixes or [int.to_bytes(i, output_size, 'little') for i in range(2 ** k)]
        self.collision_func = collision_func

        # Output fields
        self.hash_tree = {}
        self.crafted_hash = None

        self._generate_tree()



    def _generate_tree(self):
        """
        Builds a binary hash tree of colliding, intermediary Merkle-Damgard construction states.
        """
        log.debug('Generating hash tree')
        tree = []
        for i in range(self.k):
            tree.append([])


        promoted_prefix = None

        # Determine if we need to "promote" a prefix. Basically, if we have an odd number of
        # prefixes on this layer, we put the last prefix in waiting as there will eventually be
        # another layer which has one less node than a power of two.
        curr_prefix_list = self.prefixes
        if len(self.prefixes) % 2 == 1:
            promoted_prefix = self.prefixes[-1]
            curr_prefix_list = self.prefixes[:-1]

        tree[0] = [(curr_prefix_list[i], curr_prefix_list[i + 1]) for i in range(0, len(curr_prefix_list), 2)]

        solution_tree = []
        for i in range(self.k):
            solution_tree.append([])


        for i in range(self.k):
            for (p1, p2) in tree[i]:
                p1_suffix, p2_suffix, intermediary_collision = self.collision_func(p1, p2)
                solution_tree[i].append((p1, p2, p1_suffix, p2_suffix, intermediary_collision))

            # Add solutions
            if i < (self.k - 1):

                # If there's an odd number of solutions at this level, then there's either a prefix
                # in waiting, or we need to promote one.
                next_level_states = solution_tree[i]
                if len(solution_tree[i]) % 2 == 1:
                    next_level_states = deepcopy(solution_tree[i])

                    # Last level is a multiple of 2 but not a power of 2 (e.g. 6).
                    # Promote our last prefix.
                    if promoted_prefix == None:
                        promoted_prefix = next_level_states[-1][-1]
                        next_level_states = next_level_states[:-1]

                    # We have a prefix in waiting. Use it immediately.
                    else:
                        next_level_states.append((promoted_prefix,))

                tree[i + 1] = [(next_level_states[sol][-1], next_level_states[sol + 1][-1]) for sol in range(0, len(next_level_states), 2)]

        # We're done generating the tree; time to set the output fields
        for layer in solution_tree:
            for p1_init, p2_init, p1_msg, p2_msg, result in layer:
                self.hash_tree[p1_init] = (p1_init, p2_init, p1_msg, p2_msg, result)
                self.hash_tree[p2_init] = (p1_init, p2_init, p1_msg, p2_msg, result)
        
        self.crafted_hash = solution_tree[-1][0][-1]



    def execute(self, message: bytes) -> Bytes:
        """
        Traverves the hash tree and builds an appropriate suffix to cause the collision.

        Parameters:
            message (bytes): Bytes-like message to find a suffix for.
        
        Returns:
            Bytes: either an empty byte-string or a suffix that causes the collision.
        """
        suffix = b''

        while message in self.hash_tree:
            found_node = self.hash_tree[message]

            if found_node[0] == message:
                next_suffix = found_node[2]
            else:
                next_suffix = found_node[3]

            suffix += next_suffix
            message = found_node[-1]

        return Bytes(suffix)


from types import FunctionType

# https://en.wikipedia.org/wiki/Xorshift

MASK32 = 0xFFFFFFFF
MASK58 = 0x3FFFFFFFFFFFFFF
MASK64 = 0xFFFFFFFFFFFFFFFF

def V32(x):
    x ^= (x << 13) & MASK32
    x ^= x >> 17
    x ^= (x << 5) & MASK32

    return x, x


def V64(x):
    x ^= (x << 13) & MASK64
    x ^= x >> 7
    x ^= (x << 17) & MASK64

    return x, x


def V128(x):
    s = x[0]
    t = x[3]
    t ^= (t << 11) & MASK64
    t ^= t >> 8

    x[3] = x[2]
    x[2] = x[1]
    x[1] = x[0]

    t ^= s >> 19
    t ^= s
    t &= MASK64

    return [t, *x[1:]], t


def V116_PLUS(state):
    s1, s0 = state
    state[0] = s0

    s1 ^= (s1 << 24) & MASK58
    state[1] = s1 ^ s0 ^ (s1 >> 11) ^ (s0 >> 41)

    return state, (state[1] + s0) & MASK58



def V128_PLUS(state):
    s1, s0 = state
    state[0] = s0

    s1 ^= (s1 << 23) & MASK64
    s1 ^= s1 >> 17
    s1 ^= s0
    s1 ^= s0 >> 26

    state[1] = s1

    return state, sum(state) & MASK64



def V1024_STAR(state):
    p, s = state
    s0 = s[p]

    p = (p + 1) & 15
    s1 = s[p]

    s1 ^= (s1 << 31) & MASK64   
    s[p] = s1 ^ s0 ^ (s1 >> 11) ^ (s0 >> 30)
    return [p, s], (s[p] * 1181783497276652981) & MASK64



class Xorshift(object):
    def __init__(self, seed: int, variant: FunctionType=V128_PLUS):
        """
        Parameters:
            seed       (int): Initial value.
            variant   (func): Xorshift variant. Function that takes in the current state (may be a list) and returns (new_state, output).
        """
        self.state = seed
        self.variant = variant

    def __repr__(self):
        return f"<Xorshift: state={self.state}, variant={self.variant}>"

    def __str__(self):
        return self.__repr__()
    

    def generate(self) -> int:
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        self.state, result = self.variant(self.state)
        return result



from math import ceil
from samson.utilities.bytes import Bytes
from types import FunctionType

class PBKDF2(object):
    def __init__(self, hash_fn: FunctionType, desired_len: int, num_iters: int):
        """
        Parameters:
            hash_fn    (func): Function that takes in bytes and returns them hashed.
            desired_len (int): Desired output length.
            num_iters   (int): Number of iterations to perform.
        """
        self.hash_fn = hash_fn
        self.desired_len = desired_len
        self.num_iters = num_iters


    def __repr__(self):
        return f"<PBKDF2: hash_fn={self.hash_fn}, desired_len={self.desired_len}, num_iters={self.num_iters}>"

    def __str__(self):
        return self.__repr__()


    def derive(self, password: bytes, salt: bytes) -> Bytes:
        """
        Derives a key.

        Parameters:
            password (bytes): Bytes-like object to key the internal state.
            salt     (bytes): Salt to tweak the output.
        
        Returns:
            Bytes: Derived key.
        """
        hash_len = len(self.hash_fn(b'', b''))
        num_blocks = ceil(self.desired_len / hash_len)

        output = Bytes(b'')
        for i in range(1, num_blocks + 1):
            xor_sum = Bytes(b'').zfill(hash_len)
            last = salt + Bytes(i).zfill(4)

            for _ in range(self.num_iters):
                last = self.hash_fn(password, last)
                xor_sum ^= last
            
            output += xor_sum
        
        return output[:self.desired_len]


import struct
import string

import logging
log = logging.getLogger(__name__)


alphabet = (string.ascii_lowercase + string.ascii_uppercase + string.digits + '=/+' + ':;').encode()
padding_chars = ')(\{\}\\[]<%@#^`~$*!|`'[::-1].encode() + b'>.\'_,?;:+-='

class CRIMEAttack(object):
    def __init__(self, oracle, alphabet=alphabet, padding_chars=padding_chars):
        self.oracle = oracle
        self.alphabet = alphabet
        self.padding_chars = padding_chars


    
    def execute(self, known_plaintext):
        plaintext = known_plaintext

        padding = self._find_padding(plaintext)

        if padding == None:
            raise Exception("No suitable padding found")

        ctr = 0
        while True:
            log.debug('Attempt format of "{}"'.format((plaintext + b'{}' + padding).decode()))
            padded_sizes = [(struct.pack('B', char), self.oracle.request((plaintext + struct.pack('B', char)) + padding)) for char in self.alphabet]

            sorted_sizes = sorted(padded_sizes, key=lambda req: req[1])
            log.debug('Sizes for iteration {}: {}'.format(ctr, sorted_sizes))

            if sorted_sizes[0][1] == sorted_sizes[1][1]:
                return plaintext
            else:
                plaintext += sorted_sizes[0][0]

            ctr += 1




    def _find_padding(self, payload):
        reference_len = self.oracle.request(payload)
        padding = b''

        log.debug('Attempting to find padding')

        for char in self.padding_chars:
            padding += struct.pack('B', char)
            new_len = self.oracle.request(payload + padding)

            if new_len > reference_len:
                padding = padding[:-1]
                log.debug('Found suitable padding "{}"'.format(padding.decode()))
                return padding
        


from samson.utilities.bytes import Bytes
from samson.utilities.manipulation import left_rotate
from samson.constructions.merkle_damgard_construction import MerkleDamgardConstruction

# http://cacr.uwaterloo.ca/hac/about/chap9.pdf
RL = [
    [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 ],
    [ 7, 4, 13, 1, 10, 6, 15, 3, 12, 0, 9, 5, 2, 14, 11, 8 ],
    [ 3, 10, 14, 4, 9, 15, 8, 1, 2, 7, 0, 6, 13, 11, 5, 12 ],
    [ 1, 9, 11, 10, 0, 8, 12, 4, 13, 3, 7, 15, 14, 5, 6, 2 ],
    [ 4, 0, 5, 9, 7, 12, 2, 10, 14, 1, 3, 8, 11, 6, 15, 13 ]
]

RR = [
    [ 5, 14, 7, 0, 9, 2, 11, 4, 13, 6, 15, 8, 1, 10, 3, 12 ],
    [ 6, 11, 3, 7, 0, 13, 5, 10, 14, 15, 8, 12, 4, 9, 1, 2 ],
    [ 15, 5, 1, 3, 7, 14, 6, 9, 11, 8, 12, 2, 10, 0, 4, 13 ],
    [ 8, 6, 4, 1, 3, 11, 15, 0, 5, 12, 2, 13, 9, 7, 10, 14 ],
    [ 12, 15, 10, 4, 1, 5, 8, 7, 6, 2, 13, 14, 0, 3, 9, 11 ]
]


SL = [
    [ 11, 14, 15, 12, 5, 8, 7, 9, 11, 13, 14, 15, 6, 7, 9, 8 ],
    [ 7, 6, 8, 13, 11, 9, 7, 15, 7, 12, 15, 9, 11, 7, 13, 12 ],
    [ 11, 13, 6, 7, 14, 9, 13, 15, 14, 8, 13, 6, 5, 12, 7, 5 ],
    [ 11, 12, 14, 15, 14, 15, 9, 8, 9, 14, 5, 6, 8, 6, 5, 12 ],
    [ 9, 15, 5, 11, 6, 8, 13, 12, 5, 12, 13, 14, 11, 8, 5, 6 ]
]

SR = [
    [ 8, 9, 9, 11, 13, 15, 15, 5, 7, 7, 8, 11, 14, 14, 12, 6 ],
    [ 9, 13, 15, 7, 12, 8, 9, 11, 7, 7, 12, 7, 6, 15, 13, 11 ],
    [ 9, 7, 15, 11, 8, 6, 6, 14, 12, 13, 5, 14, 13, 13, 7, 5 ],
    [ 15, 5, 8, 11, 14, 14, 6, 14, 6, 9, 12, 9, 12, 5, 15, 8 ],
    [ 8, 5, 12, 9, 12, 5, 14, 6, 8, 13, 6, 5, 15, 13, 11, 11 ]
]


F1 = lambda x, y, z: x ^ y ^ z
F2 = lambda x, y, z: (x & y) | (~x & z)
F3 = lambda x, y, z: (x | ~y) ^ z
F4 = lambda x, y, z: (x & z) | (y & ~z)
F5 = lambda x, y, z: x ^ (y | ~z)

FL = [F1, F2, F3, F4, F5]
FR = [F5, F4, F3, F2, F1]

KL = [0x00000000, 0x5A827999, 0x6ED9EBA1, 0x8F1BBCDC, 0xA953FD4E]
KR = [0x50A28BE6, 0x5C4DD124, 0x6D703EF3, 0x7A6D76E9, 0x00000000]

INIT_STATE = Bytes(int.to_bytes(0x0123456789ABCDEFFEDCBA9876543210F0E1D2C3, 20, 'big'), byteorder='little')



def COMPRESS(message, state):
    # The authors of RIPEMD160 couldn't decide on whether to use big or little endian, so they used both!
    # RIPEMD160 takes in bytes as big endian but operates and outputs bytes of little endian. Man, was this 'fun.'

    h = [chunk.to_int() for chunk in state.chunk(4)]
    msg_chunks = [chunk[::-1].to_int() for chunk in Bytes.wrap(message, byteorder='big').chunk(4)]

    AL = AR = h[0]
    BL = BR = h[1]
    CL = CR = h[2]
    DL = DR = h[3]
    EL = ER = h[4]

    for curr_round in range(5):
        for w in range(16):
            T = left_rotate(AL + FL[curr_round](BL, CL, DL) + msg_chunks[RL[curr_round][w]] + KL[curr_round], SL[curr_round][w]) + EL
            AL = EL & 0xFFFFFFFF; EL = DL & 0xFFFFFFFF; DL = left_rotate(CL, 10); CL = BL & 0xFFFFFFFF; BL = T & 0xFFFFFFFF

            T = left_rotate(AR + FR[curr_round](BR, CR, DR) + msg_chunks[RR[curr_round][w]] + KR[curr_round], SR[curr_round][w]) + ER
            AR = ER & 0xFFFFFFFF; ER = DR & 0xFFFFFFFF; DR = left_rotate(CR, 10); CR = BR & 0xFFFFFFFF; BR = T & 0xFFFFFFFF
    
    T = (h[1] + CL + DR) & 0xFFFFFFFF
    h[1] = (h[2] + DL + ER) & 0xFFFFFFFF
    h[2] = (h[3] + EL + AR) & 0xFFFFFFFF
    h[3] = (h[4] + AL + BR) & 0xFFFFFFFF
    h[4] = (h[0] + BL + CR) & 0xFFFFFFFF
    h[0] = T 

    return sum([Bytes(state, 'little').zfill(4) for state in h])


class RIPEMD160(MerkleDamgardConstruction):
    """
    Stands for RACE Integrity Primitives Evaluation Message Digest (RIPEMD). While there exist other
    versions of RIPEMD (128, 256, and 320), 160 is the most popular.
    """

    def __init__(self, initial_state: bytes=INIT_STATE):
        """
        Parameters:
            initial_state (bytes): (Optional) Initial internal state.
        """
        super().__init__(
            initial_state=initial_state,
            compression_func=COMPRESS,
            digest_size=20,
            endianness='little'
        )


    def __repr__(self):
        return f"<RIPEMD160: initial_state={self.initial_state}, block_size={self.block_size}>"


    def __str__(self):
        return self.__repr__()


import os
import setuptools
from distutils.core import setup

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md')) as f:
    README = f.read()


# see requirements.txt for requirements
requires = [
    "scipy",
    "fastecdsa",
    "sympy",
    "pyasn1"
]

tests_require = [
]

setup(name='samson-crypto',
      data_files=['VERSION'],
      version=__import__('samson').VERSION,
      description='Cryptanalysis and attack framework',
      scripts=['scripts/samson'],
      long_description=README,
      classifiers=[
          "Topic :: Security",
          "Topic :: Security :: Cryptography",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3.6",
      ],
      author='Daniel Cronce',
      author_email='daniel.cronce@wildcardcorp.com',
      keywords='cryptography security cryptanalysis',
      url="https://github.com/wildcardcorp/samson",
      packages=[
          "samson",
          "samson/analyzers",
          "samson/attacks",
          "samson/auxiliary",
          "samson/block_ciphers",
          "samson/classical",
          "samson/constructions",
          "samson/kdfs",
          "samson/hashes",
          "samson/macs",
          "samson/oracles",
          "samson/padding",
          "samson/prngs",
          "samson/protocols",
          "samson/publickey",
          "samson/stream_ciphers",
          "samson/utilities"
      ],
      include_package_data=True,
      install_requires=requires)


from samson.utilities.bytes import Bytes
from types import FunctionType

class FeistelNetwork(object):
    """
    Construction for building block ciphers. Operates on half of a plaintext or ciphertext at a time and
    interleaves invocations of the Feistel function. If the Feistel function (`round_func`) is a cryptographically
    secure pseudorandom function, then three rounds are sufficient to make the cipher a psuedorandom permutation.
    Four rounds makes it a strong psuedorandom permutation.
    """

    def __init__(self, round_func: FunctionType, key_schedule: FunctionType):
        """
        Parameter:
            round_func   (func): The Feistel function that takes in a state and subkey and returns a new state. Does not need to be invertible.
            key_schedule (func): Function that takes in a key and returns a list or generator of subkeys.
        """
        self.round_func = round_func
        self.key_schedule = key_schedule


    def __repr__(self):
        return f"<FeistelNetwork round_func={self.round_func}, key_schedule={self.key_schedule}>"

    def __str__(self):
        return self.__repr__()



    def yield_encrypt(self, key: bytes, plaintext: bytes):
        """
        Yields the intermediate, encrypted states of the `plaintext`.

        Parameters:
            key       (bytes): Bytes-like object to key the cipher.
            plaintext (bytes): Bytes-like object to be encrypted.
        
        Returns:
            generator: Resulting intermediate ciphertext.
        """
        half = len(plaintext) // 2
        plaintext = Bytes.wrap(plaintext)
        L_i, R_i = plaintext[:half], plaintext[half:]
        
        round_keys = list(self.key_schedule(key))

        for subkey in round_keys:
            L_i, R_i = R_i, (L_i ^ self.round_func(R_i, subkey))
            yield R_i, L_i



    def yield_decrypt(self, key: bytes, ciphertext: bytes):
        """
        Yields the intermediate, decrypted states of the `ciphertext`.

        Parameters:
            key        (bytes): Bytes-like object to key the cipher.
            ciphertext (bytes): Bytes-like object to be decrypted.
        
        Returns:
            generator: Resulting intermediate plaintext.
        """
        half = len(ciphertext) // 2
        ciphertext = Bytes.wrap(ciphertext)
        R_i, L_i = ciphertext[:half], ciphertext[half:]

        round_keys = list(self.key_schedule(key))[::-1]

        for subkey in round_keys:
            R_i, L_i = L_i, (R_i ^ self.round_func(L_i, subkey))
            yield L_i, R_i



    def encrypt(self, key: bytes, plaintext: bytes) -> Bytes:
        """
        Encrypts `plaintext` by yielding the final state of the Feistel network.

        Parameters:
            key       (bytes): Bytes-like object to key the cipher.
            plaintext (bytes): Bytes-like object to be encrypted.
        
        Returns:
            Bytes: Resulting ciphertext.
        """
        return Bytes(b''.join(list(self.yield_encrypt(key, plaintext))[-1]))

    

    def decrypt(self, key: bytes, ciphertext: bytes) -> Bytes:
        """
        Decrypts `ciphertext` by yielding the final state of the Feistel network.

        Parameters:
        key       (bytes): Bytes-like object to key the cipher.
            ciphertext (bytes): Bytes-like object to be decrypted.
        
        Returns:
            Bytes: Resulting plaintext.
        """
        return Bytes(b''.join(list(self.yield_decrypt(key, ciphertext))[-1]))


from samson.utilities.manipulation import xor_buffs, left_rotate, right_rotate, get_blocks, transpose, stretch_key
from samson.utilities.encoding import int_to_bytes
from samson.utilities.general import rand_bytes
import codecs

class Bytes(bytearray):
    """
    Bytearray convenience class. Supports popular manipulations such as XOR, stretching, chunking, transposing, and rotations.
    Manipulations always return a new Bytes instance instead of editing the old one.
    """

    def __init__(self, bytes_like: bytes, byteorder: str='big'):
        """
        Parameters:
            bytes_like (bytes): Any bytes-like interface or integer. Integers will be converted using the byteorder.
            byteorder    (str): Byte order of the input. Will be used when converting between integer and byte representations.
        """
        if type(bytes_like) is int:
            bytes_like = int_to_bytes(bytes_like, byteorder)

        super().__init__(bytes_like)
        self.byteorder = byteorder


    @staticmethod
    def wrap(bytes_like: bytes, byteorder: str='big'):
        """
        Conditional initialization. Only creates a new Bytes object if it already isn't one.

        Parameters:
            bytes_like (bytes): Any bytes-like interface or integer. Integers will be converted using the byteorder.
            byteorder    (str): Byte order of the input. Will be used when converting between integer and byte representations.
        
        Returns:
            Bytes: A Bytes representation of the object.
        """
        if isinstance(bytes_like, Bytes):
            return bytes_like
        else:
            return Bytes(bytes_like, byteorder=byteorder)

    
    @staticmethod
    def random(size: int=16, byteorder: str='big'):
        """
        Generates a random Bytes object using /dev/urandom.

        Parameters:
            size      (int): Number of bytes to generate.
            byteorder (int): Byteorder.
        
        Returns:
            Bytes: Random Bytes.
        """
        return Bytes(rand_bytes(size), byteorder=byteorder)



    def __repr__(self):
        return f'<Bytes: {str(bytes(self))}, byteorder={self.byteorder}>'

    def __str__(self):
        return self.__repr__()


    # Operators
    def __xor__(self, other):
        if type(other) is int:
            return Bytes(int.to_bytes(self.to_int() ^ other, len(self), self.byteorder), self.byteorder)
        else:
            return Bytes(xor_buffs(self, other), self.byteorder)


    def __rxor__(self, other):
        return self.__xor__(other)


    def __getitem__(self, index):
        result = bytearray.__getitem__(self, index)
        if type(result) is int:
            return result
        else:
            return Bytes(result, self.byteorder)
        

    def __and__(self, other):
        other_as_int = other

        if not type(other) is int:
            other_as_int = int.from_bytes(other, self.byteorder)

        return Bytes(int.to_bytes(self.to_int() & other_as_int, len(self), self.byteorder), self.byteorder)


    def __rand__(self, other):
        return self.__and__(other)


    def __or__(self, other):
        other_as_int = other

        if not type(other) is int:
            other_as_int = int.from_bytes(other, self.byteorder)

        return Bytes(int.to_bytes(self.to_int() | other_as_int, len(self), self.byteorder), self.byteorder)


    def __ror__(self, other):
        return self.__or__(other)


    def __add__(self, other):
        return Bytes(bytearray.__add__(self, other), self.byteorder)


    def __radd__(self, other):
        return Bytes(bytearray(other).__add__(self), self.byteorder)


    def __lshift__(self, num):
        return Bytes(int.to_bytes((self.to_int() << num) %  (2**(len(self) * 8)), len(self), self.byteorder))


    def __rshift__(self, num):
        return Bytes(int.to_bytes((self.to_int() >> num) % (2**(len(self) * 8)), len(self), self.byteorder))


    def __invert__(self):
        max_val = 2 ** (len(self) * 8) - 1
        return Bytes(max_val - self.to_int(), self.byteorder)


    # Manipulations
    def lrot(self, amount: int, bits: int=None):
        """
        Performs a left-rotate.

        Parameters:
            amount (int): Amount to rotate by.
            bits   (int): Bitspace to rotate over.
        
        Returns:
            Bytes: A new instance of Bytes with the transformation applied.
        """
        if not bits:
            bits = len(self) * 8

        back_to_bytes = int.to_bytes(left_rotate(self.to_int(), amount, bits), bits // 8, self.byteorder)
        return Bytes(back_to_bytes, self.byteorder)



    def rrot(self, amount: int, bits: int=None):
        """
        Performs a right-rotate.

        Parameters:
            amount (int): Amount to rotate by.
            bits   (int): Bitspace to rotate over.
        
        Returns:
            Bytes: A new instance of Bytes with the transformation applied.
        """
        if not bits:
            bits = len(self) * 8

        back_to_bytes = int.to_bytes(right_rotate(self.to_int(), amount, bits), bits // 8, self.byteorder)
        return Bytes(back_to_bytes, self.byteorder)



    def chunk(self, size: int, allow_partials: bool=False) -> list:
        """
        Chunks the Bytes into `size` length chunks.

        Parameters:
            size            (int): Size of the chunks.
            allow_partials (bool): Whether or not to allow the last chunk to be a partial.
        
        Returns:
            list: List of Bytes.
        """
        return get_blocks(self, size, allow_partials)



    def transpose(self, size: int):
        """
        Builds a matrix of `size` row-length, transposes the matrix, and collapses it back into a Bytes object.

        Parameters:
            size (int): Length of the rows/chunks.
        
        Returns:
            Bytes: Transposed bytes.
        """
        return Bytes(b''.join(transpose(self, size)), self.byteorder)



    def zfill(self, size: int):
        """
        Fills the Bytes to specified size with NUL bytes _such that the integer representation stays the same according to the byteorder_.

        Parameters:
            size (int): Size of the resulting Bytes.
        
        Return:
            Bytes: Bytes padded with zeroes.
        """
        return Bytes(int.to_bytes(self.to_int(), size, self.byteorder), self.byteorder)



    def stretch(self, size: int, offset: int=0):
        """
        Repeats a Bytes object until it reaches `size` length shifted by `offset`.

        Examples:

        >>> stretch_key(b'abc', 5)
        b'abcab'

        >>> stretch_key(b'abc', 5, offset=1)
        b'cabca'


        Parameters:
            size   (int): Size to be stretched to.
            offset (int): Offset to start from.
        
        Returns:
            Bytes: Bytes stretched to `size`.
        """
        return Bytes(stretch_key(self, size, offset), self.byteorder)

    

    # Conversions
    def to_int(self) -> int:
        """
        Converts to an integer representation.

        Returns:
            int: Integer representation.
        """
        return int.from_bytes(self, self.byteorder)


    def int(self) -> int:
        """
        Converts to an integer representation.

        Returns:
            int: Integer representation.
        """
        return self.to_int()

    
    def to_hex(self):
        """
        Converts to an hex representation.

        Returns:
            Bytes: Hex representation.
        """
        return Bytes(codecs.encode(self, 'hex_codec'), self.byteorder)

    
    def hex(self):
        """
        Converts to an hex representation.
        
        Returns:
            Bytes: Hex representation.
        """
        return self.to_hex()


    def unhex(self):
        """
        Converts from an hex representation.

        Returns:
            Bytes: Raw bytes representation.
        """
        return Bytes(codecs.decode(self, 'hex_codec'), self.byteorder)


    def to_bin(self):
        """
        Converts to a Bitstring representation.

        Returns:
            Bitstring: Bitstring representation.
        """
        from samson.utilities.bitstring import Bitstring
        return Bitstring(self, byteorder=self.byteorder).zfill(len(self) * 8)

    
    def bin(self):
        """
        Converts to a Bitstring representation.

        Returns:
            Bitstring: Bitstring representation.
        """
        return self.to_bin()

    
    def to_bits(self):
        """
        Converts to a Bitstring representation.

        Returns:
            Bitstring: Bitstring representation.
        """
        return self.to_bin()
    
    
    def bits(self):
        """
        Converts to a Bitstring representation.

        Returns:
            Bitstring: Bitstring representation.
        """
        return self.to_bits()


from samson.utilities.bytes import Bytes
from samson.utilities.manipulation import left_rotate, right_rotate
import math

P_w = [0xB7E1, 0xB7E15163, 0xB7E151628AED2A6B]
Q_w = [0x9E37, 0x9E3779B9, 0x9E3779B97F4A7C15]

# https://en.wikipedia.org/wiki/RC5#Algorithm
class RC5(object):
    """
    Structure: Feistel Network
    Key size: 0-2040 bits
    Block size: 32, 64, 128 bits
    """

    def __init__(self, key: bytes, num_rounds: int=12, block_size: int=128):
        """
        Parameters:
            key      (bytes): Bytes-like object to key the cipher.
            num_rounds (int): Number of rounds to perform.
            block_size (int): The desired block size in bits.
        """
        if not block_size in [32, 64, 128]:
            raise Exception("Invalid block size: must be 32, 64, or 128 bits")

        self.key = Bytes.wrap(key)
        self.num_rounds = num_rounds
        self.block_size = block_size // 2
        self.mod = 2 ** self.block_size
        self.S = self._key_expansion()


    def __repr__(self):
        return f"<RC5: key={self.key}, num_rounds={self.num_rounds}, block_size={self.block_size}, S={self.S}>"


    def __str__(self):
        return self.__repr__()



    def _key_expansion(self):
        b = len(self.key)
        u = self.block_size // 8
        t = 2 * (self.num_rounds + 1)

        if b == 0:
            c = 1
        elif b % u:
            self.key = self.key.zfill(u - b % u + b)
            b = len(self.key)
            c = b // u
        else:
            c = b // u

        const_idx = int(math.log(self.block_size, 2) - 4)

        if b == 0:
            L = [0]
        else:
            L = self.key.chunk(b // c)

        for i in range(b - 1, -1, -1):
            L[i // u] = Bytes.wrap(L[i // u] << 8).int() + self.key[i]
        
        S = [(P_w[const_idx] + (Q_w[const_idx] * i)) % self.mod for i in range(t)]

        i = j = 0
        A = B = 0

        for _ in range(3 * max(t, c)):
            A = S[i] = left_rotate((S[i] + A + B), 3, bits=self.block_size)
            B = L[j] = left_rotate((L[j] + A + B), (A + B) % self.block_size, bits=self.block_size)
            i = (i + 1) % t
            j = (j + 1) % c
        
        return S



    def encrypt(self, plaintext: bytes) -> Bytes:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (bytes): Bytes-like object to be encrypted.
        
        Returns:
            Bytes: Resulting ciphertext.
        """
        plaintext = Bytes.wrap(plaintext)

        A = plaintext[self.block_size // 8:].int()
        B = plaintext[:self.block_size // 8].int()

        A = (A + self.S[0]) % self.mod
        B = (B + self.S[1]) % self.mod

        for i in range(1, self.num_rounds + 1):
            A = (left_rotate(A ^ B, B % self.block_size, bits=self.block_size) + self.S[2*i]) % self.mod
            B = (left_rotate(B ^ A, A % self.block_size, bits=self.block_size) + self.S[2*i + 1]) % self.mod

        return Bytes(A, 'little').zfill(self.block_size // 8) + Bytes(B, 'little').zfill(self.block_size // 8)



    def decrypt(self, ciphertext: bytes) -> Bytes:
        """
        Decrypts `ciphertext`.

        Parameters:
            ciphertext (bytes): Bytes-like object to be decrypted.
        
        Returns:
            Bytes: Resulting plaintext.
        """
        ciphertext = Bytes.wrap(ciphertext)

        A = ciphertext[:self.block_size // 8].int()
        B = ciphertext[self.block_size // 8:].int()

        for i in range(self.num_rounds, 0, -1):
            B = right_rotate((B - self.S[2*i + 1]) % self.mod, A % self.block_size, bits=self.block_size) ^ A
            A = right_rotate((A - self.S[2*i]) % self.mod, B % self.block_size, bits=self.block_size) ^ B
            
        A = (A - self.S[0]) % self.mod
        B = (B - self.S[1]) % self.mod

        return (Bytes(A, 'little').zfill(self.block_size // 8) + Bytes(B, 'little').zfill(self.block_size // 8))[::-1]


from samson.utilities.manipulation import get_blocks
from sympy import Poly, GF
from sympy.abc import x
import base64
import re

# https://en.wikipedia.org/wiki/Non-adjacent_form
def to_NAF(input_arg: bytes) -> list:
    """
    Converts bytes/bytearray or int to Non-adjacent form (NAF).

    Parameters:
        input_arg (bytes): Raw bytes/integer.
    
    Returns:
        list: Sequence in NAF.
    """
    if type(input_arg) is int:
        E = input_arg
    else:
        E = int.from_bytes(input_arg, 'big')
    
    z = []
    
    i = 0
    while E > 0:
        if E % 2 == 1:
            z.append(int(2 - (E % 4)))
            E -= z[-1]
        else:
            z.append(0)
        
        E /= 2
        i += 1
    return z[::-1]


def from_NAF(naf: list) -> int:
    """
    Converts a NAF sequence into an integer.

    Parameters:
        naf (list): NAF sequence.
    
    Returns:
        int: Integer representation.
    """
    total = 0
    reversed_naf = naf[::-1]
    for i in range(len(naf)):
        total += 2 ** i * reversed_naf[i]
    
    return total




def int_to_bytes(n: int, byteorder: str='big') -> bytes:
    """
    Converts an int `n` to bytes.

    Parameters:
        n         (int): Integer.
        byteorder (str): Desired byte order ('big' or 'little').
    
    Returns:
        bytes: Bytes representation of `n`.
    """
    return n.to_bytes(max((n.bit_length() + 7) // 8, 1), byteorder)



def bytes_to_bitstring(input_bytes: bytes, fill: int=8) -> str:
    """
    Converts bytes to a bitstring.

    Parameters:
        input_bytes (bytes): Bytes to convert.
        fill          (int): Length of the output bitstring. Pads with zeroes.
    
    Returns:
        bytes: Bytes representation of `n`.
    """
    return ''.join(format(x, 'b').zfill(fill) for x in input_bytes)


# https://stackoverflow.com/questions/32675679/convert-binary-string-to-bytearray-in-python-3
def bitstring_to_bytes(bitstring: str, byteorder: str='big') -> bytes:
    """
    Converts a bitstring to bytes.

    Parameters:
        bitstring (str): Bitstring to convert.
        byteorder (str): Desired byte order ('big' or 'little').
    
    Returns:
        bytes: Bytes representation.
    """
    return int(bitstring, 2).to_bytes(len(bitstring) // 8, byteorder=byteorder)



PRE_ENCAPSULATION_BOUNDARY_RE  = re.compile(rb'\s*-----BEGIN (.*)-----\n')
POST_ENCAPSULATION_BOUNDARY_RE = re.compile(rb'-----END (.*)-----\s*')

# TODO: Handle encrypted PEM files
def pem_decode(pem_bytes: bytes) -> bytes:
    """
    Decodes PEM bytes into raw bytes.

    Parameters:
        pem_bytes (bytes): PEM-encoded bytes.

    Returns:
        bytes: Decoded bytes.
    """
    if not PRE_ENCAPSULATION_BOUNDARY_RE.match(pem_bytes):
        raise ValueError('`pem_bytes` must have a valid pre-encapsulation boundary')


    if not POST_ENCAPSULATION_BOUNDARY_RE.search(pem_bytes):
        raise ValueError('`pem_bytes` must have a valid post-encapsulation boundary')


    headers_removed = b''.join(pem_bytes.replace(b' ', b'').split()[1:-1])
    return base64.b64decode(headers_removed)



def pem_encode(der_bytes: bytes, marker: str, width: int=70) -> bytes:
    """
    PEM-encodes DER-encoded bytes.

    Parameters:
        der_bytes (bytes): DER-encoded bytes.
        marker      (str): Header and footer marker (e.g. 'RSA PRIVATE KEY').
        width       (int): Maximum line width before newline.

    Returns:
        bytes: PEM-encoded bytes.
    """
    data = b'\n'.join(get_blocks(base64.b64encode(der_bytes), block_size=width, allow_partials=True))
    return f"-----BEGIN {marker}-----\n".encode('utf-8') + data + f"\n-----END {marker}-----".encode('utf-8')



def int_to_poly(integer: int) -> Poly:
    """
    Encodes an integer as a polynomial.

    Parameters:
        integer (int): Integer to encode.
    
    Returns:
        Poly: Polynomial representation.
    """
    return Poly(sum([int(bit) * x ** i for i, bit in enumerate(bin(integer)[2:][::-1])][::-1]) + x, domain=GF(2)) - Poly(x, domain=GF(2))



def poly_to_int(poly: Poly) -> int:
    """
    Encodes an polynomial as a integer.

    Parameters:
        poly (Poly): Polynomial to encode.
    
    Returns:
        int: Integer representation.
    """
    return int(''.join([str(bit) for bit in poly.all_coeffs()]), 2)


from samson.utilities.bytes import Bytes
from types import FunctionType

# https://en.wikipedia.org/wiki/One-way_compression_function#Davies%E2%80%93Meyer
class DaviesMeyerConstruction(object):
    """
    A Davies-Meyer construction is a one-way compression function built from a block cipher.
    """

    def __init__(self, initial_state: bytes, encryptor: FunctionType):
        """
        Parameters:
            initial_state (bytes): Bytes-like initial state that is the correct size for the underlying cipher.
            encryptor      (func): Function that takes in a plaintext and returns a ciphertext.
        """
        self.initial_state = Bytes.wrap(initial_state)
        self.block_size = len(self.initial_state)
        self.encryptor = encryptor
    

    def __repr__(self):
        return f"<DaviesMeyerConstruction initial_state={self.initial_state}, encryptor={self.encryptor}, block_size={self.block_size}>"

    def __str__(self):
        return self.__repr__()


    def yield_state(self, message: bytes):
        """
        Yields the intermediate, hashed states of the `message`.

        Parameters:
            message (bytes): Message to be hashed.
        
        Returns:
            generator: Intermediate, hashed states.
        """
        message = Bytes.wrap(message)
        last_state = self.initial_state

        for block in message.chunk(self.block_size):
            last_state = self.encryptor(block, last_state) ^ last_state
            yield last_state



    def hash(self, message: bytes) -> Bytes:
        """
        Yields the final, hashed state of the `message`.

        Parameters:
            message (bytes): Message to be hashed.
        
        Returns:
            Bytes: Fully-hashed state.
        """
        final_state = [_ for _ in self.yield_state(message)][-1]
        return final_state



    # TODO: Get block size from block_cipher object instead of having to explicitly set it.
    @staticmethod
    def generate_fixed_point(block_cipher: object, message: bytes, block_size: int):
        """
        Generates a Davies-Meyer fixed point. A fixed point is a state in which its output matches
        its input, and, therefore, infinitely produces itself.

        Parameters:
            block_cipher (object): Block cipher object.
            message       (bytes): Message you want to be fixed point.
            block_size      (int): Block size of `block_cipher`.

        Returns:
            DaviesMeyerConstruction: A DaviesMeyerConstruction with the initial state set to the fixed point.
        """
        message = Bytes.wrap(message)
        first_block = message.chunk(block_size)[0]

        initial_state = block_cipher(first_block).decrypt(Bytes(b'').zfill(block_size))

        return DaviesMeyerConstruction(initial_state=initial_state, encryptor=lambda key, msg: block_cipher(key).encrypt(msg))


from samson.utilities.bytes import Bytes
from types import FunctionType

class EvenMansourScheme(object):
    """
    Block cipher construction built from a prewhitening key, unkeyed psuedorandom permutation, and postwhitening key.
    """

    def __init__(self, F: FunctionType, K1: bytes, K2: bytes=None):
        """
        Parameters:
            F   (func): Unkeyed psuedorandom permutation.
            K1 (bytes): Bytes-like object to key the cipher.
            K2 (bytes): (Optional) Bytes-like object to key the cipher.
        """
        self.F = F
        self.K1 = Bytes.wrap(K1)
        self.K2 = Bytes.wrap(K2 or K1)
        self.block_size = len(self.K1)


    def __repr__(self):
        return f"<EvenMansourScheme F={self.F}, K1={self.K1}, K2={self.K2}, block_size={self.block_size}>"

    def __str__(self):
        return self.__repr__()
    


    def encrypt(self, plaintext: bytes) -> Bytes:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (bytes): Bytes-like object to be encrypted.
        
        Returns:
            Bytes: Resulting ciphertext.
        """
        k1_p = self.K1 ^ plaintext
        f_p = self.F(k1_p)
        return f_p ^ self.K2



    def decrypt(self, ciphertext: bytes) -> Bytes:
        """
        Decrypts `ciphertext`.

        Parameters:
            ciphertext (bytes): Bytes-like object to be decrypted.
        
        Returns:
            Bytes: Resulting plaintext.
        """
        k2_p = self.K2 ^ ciphertext
        f_p = self.F(k2_p)
        return f_p ^ self.K1


from samson.utilities.analysis import chisquare, count_items
from samson.utilities.bytes import Bytes
from samson.analyzers.english_analyzer import EnglishAnalyzer
import string


letter_freq = {
    b'e': 0.1249,
    b't': 0.0928,
    b'a': 0.0804,
    b'o': 0.0764,
    b'i': 0.0757,
    b'n': 0.0723,
    b's': 0.0651,
    b'r': 0.0628,
    b'h': 0.0505,
    b'l': 0.0407,
    b'd': 0.0382,
    b'c': 0.0334,
    b'u': 0.0273,
    b'm': 0.0251,
    b'f': 0.0240,
    b'p': 0.0214,
    b'g': 0.0187,
    b'w': 0.0168,
    b'y': 0.0166,
    b'b': 0.0148,
    b'v': 0.0105,
    b'k': 0.0054,
    b'x': 0.0023,
    b'j': 0.0016,
    b'q': 0.0012,
    b'z': 0.0009
}


letter_distribution = {}
for key, val in letter_freq.items():
    letter_distribution[ord(key)] = val



class Vigenere(object):
    """
    Polyalphabetic subsitution cipher that can be reduced to interwoven Caesar ciphers.
    """

    def __init__(self, key: bytes, alphabet=bytes(string.ascii_lowercase, 'utf-8')):
        """
        Parameters:
            key      (bytes): Bytes-like object to key the cipher.
            alphabet (bytes): Alphabet (in order) to encrypt over. Input must also be in this alphabet.
        """
        self.key = key
        self.alphabet = alphabet

    
    def __repr__(self):
        return "<Vigenere: key={}, alphabet={}>".format(self.key, self.alphabet)

    def __str__(self):
        return self.__repr__()


    def encrypt(self, plaintext: bytes) -> Bytes:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (bytes): Bytes-like object to be encrypted.
        
        Returns:
            Bytes: Resulting ciphertext.
        """
        result = []

        for i, char in enumerate(plaintext):
            k_i = bytes([self.key[i % len(self.key)]])
            c_i_idx = (self.alphabet.index(bytes([char])) + self.alphabet.index(k_i)) % len(self.alphabet)
            result.append(self.alphabet[c_i_idx])

        return Bytes(bytes(result))

    
    def decrypt(self, ciphertext: bytes) -> Bytes:
        """
        Decrypts `ciphertext`.

        Parameters:
            ciphertext (bytes): Bytes-like object to be decrypted.
        
        Returns:
            Bytes: Resulting plaintext.
        """
        result = []

        for i, char in enumerate(ciphertext):
            k_i = bytes([self.key[i % len(self.key)]])
            p_i_idx = (self.alphabet.index(bytes([char])) - self.alphabet.index(k_i)) % len(self.alphabet)
            result.append(self.alphabet[p_i_idx])

        return Bytes(bytes(result))


    @staticmethod
    def break_vigenere(ciphertext: str, alphabet: bytes=bytes(string.ascii_lowercase, 'utf-8'), expected_distribution=letter_distribution, min_key_length: int=1, max_key_length: int=20):
        ciphertext = Bytes.wrap(ciphertext)
        cipher_scores = []
        cipher_len = len(ciphertext)

        analyzer = EnglishAnalyzer()

        for i in range(min_key_length, max_key_length):
            transposed = ciphertext.transpose(i)
            total_key_score = 1
            top_chunk_scores = []

            for chunk in transposed.chunk(cipher_len // i):
                curr_chunk_scores = []

                for char in alphabet:
                    tmp_vig = Vigenere(bytes([char]))
                    new_chunk = tmp_vig.decrypt(chunk)

                    chunk_score = chisquare(count_items(new_chunk), expected_distribution)
                    curr_chunk_scores.append((char, chunk_score))


                top_chunk_scores.append(sorted(curr_chunk_scores, key=lambda chunk_score: chunk_score[1], reverse=False)[0])

            for chunk_score in top_chunk_scores:
                total_key_score += chunk_score[1]
                
            cipher_scores.append((total_key_score, top_chunk_scores))

        analyzer_scores = []
        for _total_score, top_chunk_scores in cipher_scores:
            vig = Vigenere(bytes([char for char,score in top_chunk_scores]), alphabet=alphabet)
            analyzer_scores.append((analyzer.analyze(vig.decrypt(ciphertext)) / len(vig.key), vig))

        top_analyzed_score = sorted(analyzer_scores, key=lambda kv: kv[0], reverse=True)[0]
        return top_analyzed_score[1]


from samson.utilities.math import gcd, mod_inv, lll, is_power_of_two
from sympy.matrices import Matrix
from sympy import isprime
from sympy.ntheory import factorint
from sympy.ntheory.residue_ntheory import is_primitive_root
import functools


class LCG(object):
    """
    Linear congruential generator of the form `(a*X + c) mod m`.
    """

    def __init__(self, X: int, a: int, c: int, m: int):
        """
        Parameters:
            X (int): Initial state.
            a (int): Multiplier.
            c (int): Increment.
            m (int): Modulus.
        """
        self.a = a
        self.c = c
        self.m = m

        self.X = X


    def __repr__(self):
        return f"<LCG: X={self.X}, a={self.a}, c={self.c}, m={self.m}>"


    def __str__(self):
        return self.__repr__()
        

    def generate(self) -> int:
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        self.X = (self.a * self.X + self.c) % self.m
        return self.X

    
    # https://en.wikipedia.org/wiki/Linear_congruential_generator#Period_length
    def check_full_period(self) -> bool:
        """
        Checks whether the LCG will achieve a full period with its current parameters.

        Returns:
            bool: Whether or not it will acheive a full period.
        """
        if isprime(self.m) and self.c == 0 and is_primitive_root(self.a, self.m):
            return True
        elif is_power_of_two(self.m) and self.c == 0:
            return False
        else:
            divisible_by_four = True
            if self.m % 4 == 0:
                divisible_by_four = (self.a - 1) % 4 == 0

            factors = [factor for factor, count in factorint(self.m).items()]
            divisible_by_all_factors = all([((self.a - 1) % factor) == 0 for factor in factors])
            return gcd(self.m, self.c) == 1 and divisible_by_four and divisible_by_all_factors



    # https://tailcall.net/blog/cracking-randomness-lcgs/
    @staticmethod
    def crack(states: list, multiplier: int=None, increment: int=None, modulus: int=None):
        """
        Given a few full states (probably under ten) and any (or even none) of the parameters of an LCG, returns a replica LCG.

        Parameters:
            states    (list): List of full-state outputs (in order).
            multiplier (int): (Optional) The LCG's multiplier.
            increment  (int): (Optional) The LCG's increment.
            modulus    (int): (Optional) The LCG's modulus.
        
        Returns:
            LCG: Replica LCG that predicts all future outputs of the original.
        """
        if not modulus:
            diffs = [state1 - state0 for state0, state1 in zip(states, states[1:])]
            congruences = [t2*t0 - t1*t1 for t0, t1, t2 in zip(diffs, diffs[1:], diffs[2:])]
            modulus = abs(functools.reduce(gcd, congruences))

        
        if not multiplier:
            multiplier = (states[2] - states[1]) * mod_inv(states[1] - states[0], modulus) % modulus


        if not increment:
            increment = (states[1] - states[0] * multiplier) % modulus

        return LCG(states[-1], multiplier, increment, modulus)
    
    

    # Reference: https://github.com/mariuslp/PCG_attack
    # Reference: https://www.math.cmu.edu/~af1p/Texfiles/RECONTRUNC.pdf
    # ^^ "Reconstructing Truncated Integer Variables Satisfying Linear Congruences"
    @staticmethod
    def crack_truncated(outputs: list, outputs_to_predict: list, multiplier: int, increment: int, modulus: int, trunc_amount: int):
        """
        Given a decent number of truncated states (about 200 when there's only 3-bit outputs), returns a replica LCG.

        Parameters:
            outputs            (list): List of truncated-state outputs (in order).
            outputs_to_predict (list): Next few outputs to compare against. Accuracy/number of samples trade-off.
            multiplier          (int): The LCG's multiplier.
            increment           (int): The LCG's increment.
            modulus             (int): The LCG's modulus.
        
        Returns:
            LCG: Replica LCG that predicts all future outputs of the original.
        """
        # Trivial case
        if increment == 0:
            computed_seed = LCG.solve_tlcg(outputs + outputs_to_predict, multiplier, modulus, trunc_amount)

            # Here we take the second to last seed since our implementation edits the state BEFORE it returns
            return LCG((multiplier * computed_seed[-2]) % modulus, multiplier, increment, modulus)

        else:
            diffs = [o2 - o1 for o1, o2 in zip(outputs, outputs[1:])]
            seed_diffs = LCG.solve_tlcg(diffs, multiplier, modulus, trunc_amount)
            seed_diffs = [seed_diff % modulus for seed_diff in seed_diffs]


            # Bruteforce low bits
            for z in range(2 ** trunc_amount):
                x_0 = (outputs[0] << trunc_amount) + z
                x_1 = (seed_diffs[0] + x_0) % modulus
                computed_c = (x_1 - multiplier * x_0) % modulus

                computed_x_2 = (multiplier * x_1 + computed_c) % modulus
                actual_x_2 = (seed_diffs[1] + x_1) % modulus

                if computed_x_2 == actual_x_2:
                    computed_seeds = [x_0]

                    for diff in seed_diffs:
                        computed_seeds.append((diff + computed_seeds[-1]) % modulus)
                    

                    # It's possible to find a spectrum of nearly-equivalent LCGs.
                    # The accuracy of `predicted_lcg` is dependent on the size of `outputs_to_predict` and the
                    # parameters of the LCG.
                    predicted_seed = (multiplier * computed_seeds[-2] + computed_c) % modulus
                    predicted_lcg = LCG(X=int(predicted_seed), a=multiplier, c=int(computed_c), m=modulus)

                    if [predicted_lcg.generate() >> trunc_amount for _ in range(len(outputs_to_predict))] == outputs_to_predict:
                        return predicted_lcg




    @staticmethod
    def solve_tlcg(outputs: list, multiplier: int, modulus: int, trunc_amount: int) -> Matrix:
        """
        Used internally by `crack_truncated`. Uses the LLL algorithm to find seed differentials.

        Parameters:
            outputs            (list): List of truncated-state outputs (in order).
            multiplier          (int): The LCG's multiplier.
            increment           (int): The LCG's increment.
            modulus             (int): The LCG's modulus.
        
        Returns:
            Matrix: `sympy` `Matrix` representing seed differentials.

        """
        # Initialize matrix `L`
        l_matrix = [[0 for _ in range(len(outputs))] for _ in range(len(outputs))]
        l_matrix[0][0] = modulus

        for i in range(1, len(outputs)):
            l_matrix[i][0] = multiplier ** i
            l_matrix[i][i] = -1


        l_matrix = Matrix(l_matrix)
        reduced_basis = lll([l_matrix.row(row) for row in range(l_matrix.rows)])

        # Construct and reduce `y` vector
        y = Matrix([2**trunc_amount * x % modulus for x in outputs])
        reduced_outputs = reduced_basis * y

        c_prime = Matrix([(round(x / modulus) * modulus) - x for x in reduced_outputs])
        z = Matrix(reduced_basis.LUsolve(c_prime))

        return y + z



from samson.utilities.bytes import Bytes
from samson.oracles.padding_oracle import PaddingOracle
import struct

import logging
log = logging.getLogger(__name__)

# https://grymoire.wordpress.com/2014/12/05/cbc-padding-oracle-attacks-simplified-key-concepts-and-pitfalls/
class CBCPaddingOracleAttack(object):
    """
    Performs a CBC padding oracle.

    Currently only works with PKCS7.

    Conditions:
        * CBC is being used
        * The system leaks whether the plaintext's padding was correct or not
        * The user has access to an oracle that attempts to decrypt arbitrary ciphertext
    """
    def __init__(self, oracle: PaddingOracle, iv: bytes, block_size: int=16):
        """
        Parameters:
            oracle (PaddingOracle): An oracle that takes in a bytes-like object and returns a boolean
                                    indicating whether the padding was correct.
            iv             (bytes): Initialization vector (or previous ciphertext block) of the ciphertext
                                    to crack.
            block_size       (int): Block size of the block cipher being used.
        """
        self.oracle = oracle
        self.iv = Bytes.wrap(iv)
        self.block_size = block_size


    def execute(self, ciphertext: bytes) -> Bytes:
        """
        Executes the attack.

        Parameters:
            ciphertext (bytes): Bytes-like ciphertext to be decrypted.

        Returns:
            Bytes: Plaintext corresponding to the inputted ciphertext.
        """
        blocks = Bytes.wrap(ciphertext).chunk(self.block_size)
        reversed_blocks = blocks[::-1]

        plaintexts = []

        for i, block in enumerate(reversed_blocks):
            log.debug("Starting iteration {}".format(i))
            plaintext = Bytes(b'')

            if i == len(reversed_blocks) - 1:
                preceding_block = self.iv
            else:
                preceding_block = reversed_blocks[i + 1]

            for _ in range(len(block)):
                last_working_char = None

                for possible_char in range(256):
                    test_byte = struct.pack('B', possible_char)
                    payload = test_byte + plaintext
                    prefix = b'\x00' * (self.block_size - len(payload))

                    padding = (struct.pack('B', len(payload)) * (len(payload))) ^ payload

                    fake_block = prefix + padding
                    exploit_block = fake_block ^ preceding_block
                    new_cipher = exploit_block + block

                    if self.oracle.check_padding(bytes(new_cipher)):
                        log.debug("Found working byte: {}".format(test_byte))
                        last_working_char = test_byte

                    # Early out optimization. Note, we're being careful about PKCS7 padding here.
                    if last_working_char and possible_char >= self.block_size:
                        break

                plaintext = last_working_char + plaintext

            plaintexts.append(plaintext)
        return Bytes(b''.join(plaintexts[::-1]))